"""
Abstractions for defining closures.

A closure is a a set of physical equations which determines the sub-mesh turbulence of the water
column at a specific time. The purpose of Tunax is to calibrate the parameters of these equations.
This module contains the abstract classes required to define a closure, which is done in the folder
:code:`closures/`. These classes can be obtained by the prefix :code:`tunax.closure.` or directly by
:code:`tunax.`.

"""

from abc import ABC
from typing import Callable, Type, TypeVar, Generic

import equinox as eqx

from tunax.case import CaseTracable
from tunax.space import Grid, State, ArrNzp1


class ClosureParametersAbstract(eqx.Module, ABC):
    """
    Abstraction for the parameters of a closure.
    
    To define a closure, a class that inherits from this one must be created. This parent class does
    not impose anything on the child one. The child class should be fill of the parameters involved
    in the closure which may be calibrated. The child class must be the only place where these
    parameters are defined for running a forward model : the instance of this class is given to the
    closure functions to recover the parameters and do the computation. For a run of a calibration,
    the description of the parameters to calibration is done from the child class. It can also
    includes parameters that are not dedicated to be calibrated such as physical or mathematical
    constants. The attributes of this class can be used to avoid the systematic computation of some
    values which are independent of the time (eg. k-epsilon with the method :code:`__post_init__`). 
    
    """


# variable that represent a type which contains ClosureParametersAbstract and
# all its subclasses
CloParT = TypeVar('CloParT', bound=ClosureParametersAbstract)


class ClosureStateAbstract(eqx.Module, ABC):
    r"""
    Abstraction for the water column state linked to the closure.

    To define a closure, a class that inherits rom this one must be created. This parent class
    imposes a grid and three variables as attributes : these variables are used at each step of the
    forward model to compute the next step of tracers and momentum. The child class should be fill
    with other arrays and scalars that may evolve throught time and necessary for the computation of
    the closure at the next step. The child class is similary to the :class:`~space.State` class but
    for the diffusivity part. The :code:`__init__` method can be overwrited to fill the variables
    with initial values. The constructor takes a grid a set of the closure parameters to initialize
    the state.

    Parameters
    ----------
    grid : Grid
        cf. :attr:`grid`.
    closure_parameters : ClosureParametersAbstract
        A set of parameters of the closure used to initialize the state.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column, should be the same than for the :class:`~space.State` instance
        used in the model.
    akt : float :class:`~jax.Array` of shape (nz+1)
        Eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : float :class:`~jax.Array` of shape (nz+1)
        Eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.

    """

    grid: Grid
    akt: ArrNzp1
    akv: ArrNzp1

    def __init__(self, grid: Grid, closure_parameters: ClosureParametersAbstract):
        pass

# variable that represent a type which contains ClosureStateAbstract and
# all its subclasses
CloStateT = TypeVar('CloStateT', bound=ClosureStateAbstract)


class Closure(eqx.Module, Generic[CloStateT, CloParT]):
    r"""
    Implementation of a physical closure for computing eddy-diffusivity.

    This class contains the three parts that defines a closure of a vertical physics. These parts
    will be used in the model to compute the eddy-diffusivity and eddy-viscosity at each time-step
    from the current water column :class:`~space.State`, the current water column state of the
    closure (child class of :class:`~ClosureStateAbstract`), the physical
    :class:`~case.CaseTracable` and the closure parameters (child class of
    :class:`~ClosureParametersAbstract`). The constructor takes all the attributes as parameters.

    Attributes
    ----------
    name : str
        The name of the closure.
    parameters_class : Type[ClosureParametersAbstract]
        A child class of :class:`~ClosureParametersAbstract` that defines the constant parameters
        used in the computation done by the closure, it includes the parameters that may be
        calibrated.
    state_class : Type[ClosureStateAbstract]
        A child class of :class:`~ClosureStateAbstract` that defines the state of the water column
        for the variables used by the closure computation.
    step_fun : Callable[[State, CloStateT, float, CloParT, CaseTracable], CloStateT]
        Ths function is called at every step of the forward model to compute the eddy-diffusivity
        before resolving the equation of the tracers and of the momentum.
        
        Parameters
        ----------
        state : State
            Current state of the water column.
        closure_state : CloStateT
            Current state of the water column for the variables used by the closure.
        dt : float
            Time-step of the forward model :math:`[\text s]`.
        closure_params : CloParT
            Values of the parameters used by the closure (time-independant).
        case_tracable : CaseTracable
            Physical parameters and forcings of the model run.
        
        Returns
        -------
        closure_state : CloStateT
            State of the water column for the variables used by the closure at the next time-step.
        
        Notes
        -----
        :code:`CloStateT` is the type reprensenting the instances of the child classes of
        :class:`~ClosureStateAbstract` and :code:`CloParT` is the same for
        :class:`~ClosureParametersAbstract`.

    """

    name : str
    parameters_class: Type[ClosureParametersAbstract]
    state_class: Type[ClosureStateAbstract]
    step_fun: Callable[[State, CloStateT, float, CloParT, CaseTracable], CloStateT]



r"""
:math:`k-\varepsilon` closure parameters, states and computation functions.

This module contains the implementation of the :math:`k-\varepsilon` model described as a GLS
case as in [1]_ as a :class:`~closure.Closure` instance. The model was traduced from Frotran to JAX
with the work of Florian Lemarié and Manolis Perrot [2]_, the translation was done in part using the
work of Anthony Zhou, Linnia Hawkins and Pierre Gentine [3]_. The parameters of the closure are
available in the :class:`KepsParameters` class, the closure state in :class:`KepsState` class. The
function :attr:`keps_step` compute one time-step of the closure, which means that it computes the
eddy-diffusivity and viscosity. The module contains other functions that are used by this main one.
These classes and the function step can be obtained by the prefix :code:`tunax.closures.k_epsilon`
or directly by :code:`tunax.closures`.

References
----------
.. [1] L. Umlauf and H. Burchard. A generic length-scale equation for geophysical turbulence models
    (2003). Journal of Marine Research 61 pp. 235-265. doi : `10.1357/002224003322005087
    <https://www.semanticscholar.org/paper/
    A-generic-length-scale-equation-for-geophysical-Umlauf-Burchard/
    24fd6403615fc7a6c5d9b6156e4f1e8d4d280af2>`_.
.. [2] M. Perrot and F. Lemarié. Energetically consistent Eddy-Diffusivity Mass-Flux convective
    schemes. Part I: Theory and Models (2024). url : `hal.science/hal-04439113
    <https://hal.science/hal-04439113>`_.
.. [3] A. Zhou, L. Hawkins and P. Gentine. Proof-of-concept: Using ChatGPT to Translate and
    Modernize an Earth System Model from Fortran to Python/JAX (2024). url :
    `arxiv.org/abs/2405.00018 <https://arxiv.org/abs/2405.00018>`_.
    
"""

from __future__ import annotations
from typing import Tuple

import equinox as eqx
import jax.numpy as jnp
from jax import lax

from tunax.case import CaseTracable
from tunax.space import Grid, State, ArrNz, ArrNzp1
from tunax.functions import FloatJax, tridiag_solve, add_boundaries
from tunax.closure import ClosureParametersAbstract, ClosureStateAbstract


class KepsParameters(ClosureParametersAbstract):
    r"""
    Parameters and constants for :math:`k-\varepsilon`.

    The first 19 attributes are the parameters of :math:`k-\varepsilon` that may be calibrated.
    This class also contains some physical constants used in the closure computing. The next 16
    attirbutes are some physical parameters for k-epsilon that can be modified but that are not
    specially supposed to be modified. The last 19 attributes are the one for the stability function
    that are computed from the parameters of :math:`k-\varepsilon`. The constructor of the class
    takes as parameters the 19 parameters of the closure and the 16 physical parameters, but not the
    last 19 stability functions for parameters.

    Attributes
    ----------
    c1 : float, default=5.
        :math:`k-\varepsilon` parameter :math:`c_1` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c2 : float, default=0.8
        :math:`k-\varepsilon` parameter :math:`c_2` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c3 : float, default=1.968
        :math:`k-\varepsilon` parameter :math:`c_3` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c4 : float, default=1.136
        :math:`k-\varepsilon` parameter :math:`c_4` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c5 : float, default=0.
        :math:`k-\varepsilon` parameter :math:`c_5` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c6 : float, default=0.4
        :math:`k-\varepsilon` parameter :math:`c_6` for the dissipation of the corelation tensor
        pressure/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cb1 : float, default=5.95
        :math:`k-\varepsilon` parameter :math:`c_{b1}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cb2 : float, default=.6
        :math:`k-\varepsilon` parameter :math:`c_{b2}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cb3 : float, default=1.
        :math:`k-\varepsilon` parameter :math:`c_{b3}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cb4 : float, default=0.
        :math:`k-\varepsilon` parameter :math:`c_{b4}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cb5 : float, default=0.3333
        :math:`k-\varepsilon` parameter :math:`c_{b5}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    cbb : float, default=.72
        :math:`k-\varepsilon` parameter :math:`c_{bb}` for the dissipation of the corelation tensor
        buoyancy/velocity (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c_mu0 : float, default=0.5477
        :math:`k-\varepsilon` parameter :math:`c_\mu^0` which links the mixing length to the
        dissipation (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    sig_k : float, default=1.
        :math:`k-\varepsilon` parameter :math:`\sigma_k` Schmit number for the dissipation of TKE
        (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    sig_eps : float, default=1.3
        :math:`k-\varepsilon` parameter :math:`\sigma_\varepsilon` Schmit number for the dissipation
        of :math:`\varepsilon` (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c_eps1 : float, default=1.44
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 1}` correction of the
        :math:`\varepsilon` equation (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c_eps2 : float, default=1.92
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 2}` correction of the
        :math:`\varepsilon` equation (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c_eps3m : float, default=-0.4
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 3}^-` correction of the
        :math:`\varepsilon` equation (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    c_eps3p : float, default=1.
        :math:`k-\varepsilon` parameter :math:`c_{\varepsilon 3}^+` correction of the
        :math:`\varepsilon` equation (Umlauf and Burchard notations) :math:`[\text{dimensionless}]`.
    chk_grav : float, default=1400.
        Charnock coefficient times gravity :math:`[\text{dimensionless}]`.
    galp: float, default=0.53
        Parameter for Galperin mixing length limitation :math:`[\text{dimensionless}]`.
    z0s_min : float, default=1e-2
        Minimal surface roughness length :math:`[\text m]`.
    z0b_min : float, default=1e-4
        Minimal bottom roughness length :math:`[\text m]`.
    z0b : float, default=1e-14
        Bottom roughness length :math:`[\text m]`.
    akt_min : float, default=1e-5
        Minimal and initialization value of eddy-diffusivity
        :math:`\left[\text m^2 \cdot \text s^{-1} \right]`.
    akv_min : float, default=1e-4
        Minimal and initialization value of eddy-viscosity
        :math:`\left[\text m^2 \cdot \text s^{-1} \right]`.
    tke_min : float, default=1e-6
        Minimal and initialization value of turbulent kinetic energy (TKE)
        :math:`\left[\text m^3 \cdot \text s^{-2} \right]`.
    eps_min : float, default=1e-12
        Minimal and initialization value of TKE dissipation
        :math:`\left[\text m^2 \cdot \text s^{-3} \right]`.
    c_mu_min : float, default=0.1
        Minimal and initialization value of :math:`c_\mu` in GLS formalisim
        :math:`[\text{dimensionless}]`.
    c_mu_prim_min : float, default=0.1
        Minimal and initialization value of `c_\mu'` in GLS formalisim
        :math:`[\text{dimensionless}]`.
    dir_sfc: bool, default=False
        Apply a Dirichlet boundary condition at the surface for TKE, else
        apply a Neumann boundary condition.
    dir_btm: bool, default=True
        Apply a Dirichlet boundary condition at the bottom for TKE, else
        apply a Neumann boundary condition.
    gls_p : float, default=3
        GLS coefficient :math:`p` to define :math:`k-\varepsilon` :math:`[\text{dimensionless}]`.
    gls_m : float, default=1.5
        GLS coefficient :math:`m` to define :math:`k-\varepsilon` :math:`[\text{dimensionless}]`.
    gls_n : float, default=-1
        GLS coefficient :math:`n` to define :math:`k-\varepsilon` :math:`[\text{dimensionless}]`.
    sf_d0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_d1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_d2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_d3 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_d4 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_d5 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_n0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_n1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_n2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_nb0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_nb1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    sf_nb2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am0 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am1 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am2 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am3 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am4 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am5 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.
    lim_am6 : float (not a parameter, computed from the above attributes)
        Limitation coefficient for :math:`k-\varepsilon` computed from the parameters
        :math:`[\text{dimensionless}]`.

    """

    # k-epsilon coefficients (Umlauf and Burchard notations)
    c1: float = 5.
    c2: float = .8
    c3: float = 1.968
    c4: float = 1.136
    c5: float = 0.
    c6: float = .4
    cb1: float = 5.95
    cb2: float = .6
    cb3: float = 1.
    cb4: float = 0.
    cb5: float = 0.3333
    cbb: float = 0.72
    c_mu0: float = 0.5477
    sig_k: float = 1.
    sig_eps: float = 1.3
    c_eps1: float = 1.44
    c_eps2: float = 1.92
    c_eps3m: float = -.4
    c_eps3p: float = 1.
    # physical constants
    chk_grav: float = 1400.
    galp: float = .53
    z0s_min: float = 1e-2
    z0b_min: float = 1e-2
    z0b: float = 1e-14
    akt_min: float = 1e-5
    akv_min: float = 1e-4
    tke_min: float = 1e-6
    eps_min: float = 1e-12
    c_mu_min: float = .1
    c_mu_prim_min: float = .1
    # physical case_tracable
    dir_sfc: bool = False
    dir_btm: bool = True
    # GLS coefficient for k-epsilon
    gls_p: float = 3
    gls_m: float = 1.5
    gls_n: float = -1
    # limitation coefficients computed from k-epsilon coefficients
    sf_d0: float = eqx.field(init=False)
    sf_d1: float = eqx.field(init=False)
    sf_d2: float = eqx.field(init=False)
    sf_d3: float = eqx.field(init=False)
    sf_d4: float = eqx.field(init=False)
    sf_d5: float = eqx.field(init=False)
    sf_n0: float = eqx.field(init=False)
    sf_n1: float = eqx.field(init=False)
    sf_n2: float = eqx.field(init=False)
    sf_nb0: float = eqx.field(init=False)
    sf_nb1: float = eqx.field(init=False)
    sf_nb2: float = eqx.field(init=False)
    lim_am0: float = eqx.field(init=False)
    lim_am1: float = eqx.field(init=False)
    lim_am2: float = eqx.field(init=False)
    lim_am3: float = eqx.field(init=False)
    lim_am4: float = eqx.field(init=False)
    lim_am5: float = eqx.field(init=False)
    lim_am6: float = eqx.field(init=False)

    def __post_init__(self):
        # stability function coefficients
        a1 = .66666666667 - .5*self.c2
        a2 = 1 - .5*self.c3
        a3 = 1 - .5*self.c4
        a5 = .5 - .5*self.c6
        nn = .5*self.c1
        nb = self.cb1
        ab1 = 1 - self.cb2
        ab2 = 1 - self.cb3
        ab3 = 2*(1 - self.cb4)
        ab5 = 2*self.cbb*(1-self.cb5)
        sf_d0 = 36*nn*nn*nn*nb*nb
        sf_d1 = 84*a5*ab3*nn*nn*nb + 36*ab5*nn*nn*nn*nb
        sf_d2 = 9*(ab2*ab2 - ab1*ab1)*nn*nn*nn - 12*(a2*a2 - 3*a3*a3)*nn*nb*nb
        sf_d3 = 12*a5*ab3*(a2*ab1 - 3*a3*ab2)*nn + 12*a5*ab3*(a3*a3 - a2*a2)*nb + \
            12*ab5*(3*a3*a3 - a2*a2)*nn*nb
        sf_d4 = 48*a5*a5*ab3*ab3*nn + 36*a5*ab3*ab5*nn*nn
        sf_d5 = 3*(a2*a2 - 3*a3*a3)*(ab1*ab1 - ab2*ab2)*nn
        sf_n0 = 36*a1*nn*nn*nb*nb
        sf_n1 = -12*a5*ab3*(ab1+ab2)*nn*nn + 8*a5*ab3*(6*a1-a2-3*a3)*nn*nb + 36*a1*ab5*nn*nn*nb
        sf_n2 = 9*a1*(ab2*ab2 - ab1*ab1)*nn*nn
        sf_nb0 = 12*ab3*nn*nn*nn*nb
        sf_nb1 = 12*a5*ab3*ab3*nn*nn
        sf_nb2 = 9*a1*ab3*(ab1 - ab2)*nn*nn + (6*a1*(a2 - 3*a3) - 4*(a2*a2 - 3*a3*a3))*ab3*nn*nb
        lim_am0 = sf_d0*sf_n0
        lim_am1 = sf_d0*sf_n1 + sf_d1*sf_n0
        lim_am2 = sf_d1*sf_n1 + sf_d4*sf_n0
        lim_am3 = sf_d4*sf_n1
        lim_am4 = sf_d2*sf_n0
        lim_am5 = sf_d2*sf_n1 + sf_d3*sf_n0
        lim_am6 = sf_d3*sf_n1
        object.__setattr__(self, 'sf_d0', sf_d0)
        object.__setattr__(self, 'sf_d1', sf_d1)
        object.__setattr__(self, 'sf_d2', sf_d2)
        object.__setattr__(self, 'sf_d3', sf_d3)
        object.__setattr__(self, 'sf_d4', sf_d4)
        object.__setattr__(self, 'sf_d5', sf_d5)
        object.__setattr__(self, 'sf_n0', sf_n0)
        object.__setattr__(self, 'sf_n1', sf_n1)
        object.__setattr__(self, 'sf_n2', sf_n2)
        object.__setattr__(self, 'sf_nb0', sf_nb0)
        object.__setattr__(self, 'sf_nb1', sf_nb1)
        object.__setattr__(self, 'sf_nb2', sf_nb2)
        object.__setattr__(self, 'lim_am0', lim_am0)
        object.__setattr__(self, 'lim_am1', lim_am1)
        object.__setattr__(self, 'lim_am2', lim_am2)
        object.__setattr__(self, 'lim_am3', lim_am3)
        object.__setattr__(self, 'lim_am4', lim_am4)
        object.__setattr__(self, 'lim_am5', lim_am5)
        object.__setattr__(self, 'lim_am6', lim_am6)


class KepsState(ClosureStateAbstract):
    r"""
    Define the state of the water column for the :math:`k-\varepsilon` model.

    The first initilisation is done from the minimal values of the different variables given in an
    instance of :class:'KepsParameters`.

    Parameters
    ----------
    grid : Grid
        cf. attribute.
    keps_params : KepsParameters
        Used to define the initialization values of the variables.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column, should be the same than for the :class:`~space.State` instance
        used in the model.
    akt : float :class:`~jax.Array` of shape (nz+1)
        Eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : float :class:`~jax.Array` of shape (nz+1)
        Eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    tke :  float :class:`~jax.Array` of shape (nz+1)
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps :  float :class:`~jax.Array` of shape (nz+1)
        TKE dissipation denoted :math:`\varepsilon` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu :  float :class:`~jax.Array` of shape (nz+1)
        :math:`c_\mu` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    c_mu_prim :  float :class:`~jax.Array` of shape (nz+1)
        :math:`c_\mu'` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.

    """

    grid: Grid
    akt: ArrNzp1
    akv: ArrNzp1
    tke: ArrNzp1
    eps: ArrNzp1
    c_mu: ArrNzp1
    c_mu_prim: ArrNzp1

    def __init__(self, grid: Grid, keps_params: KepsParameters) -> None:
        self.grid = grid
        nz = grid.nz
        self.akt = jnp.full(nz+1, keps_params.akt_min)
        self.akv = jnp.full(nz+1, keps_params.akv_min)
        self.tke = jnp.full(nz+1, keps_params.tke_min)
        self.eps = jnp.full(nz+1, keps_params.eps_min)
        self.c_mu = jnp.full(nz+1, keps_params.c_mu_min)
        self.c_mu_prim = jnp.full(nz+1, keps_params.c_mu_prim_min)


def keps_step(
        state: State,
        keps_state: KepsState,
        dt: float,
        keps_params: KepsParameters,
        case_tracable: CaseTracable
    ) -> KepsState:
    r"""
    Run one time-step of the :math:`k-\varepsilon` closure.

    The purpose of this function is to get the eddy-diffusivity and eddy-viscosity at the next time-
    step. It works in 3 steps

    1. The Brunt–Väisälä frequency and the shear is computed from the :code:`state` and the boundary
       conditions are computed.
    2. The equations on :math:`k` and :math:`\varepsilon` are solved and their values are computed
       for the next time step.
    3. The eddy-diffusivity and viscosity are computed as diagnostic variables and the
       :math:`keps_state` is updated.

    Parameters
    ----------
    state : State
        Current state of the water column.
    keps_state : KepsState
        Current state of the water column for the variables used by :math:`k-\varepsilon`.
    dt : float
        Time-step of the forward model :math:`[\text s]`.
    keps_params: KepsParameters
        Values of the parameters used by :math:`k-\varepsilon` (time-independant).
    case_tracable : CaseTracable
        Physical parameters and forcings of the model run.
    
    Returns
    -------
    keps_state : KepsState
        State of the water column for the variables used by :math:`k-\varepsilon` at the next
        time-step.
    """
    akt = keps_state.akt
    akv = keps_state.akv
    tke = keps_state.tke
    eps = keps_state.eps
    c_mu = keps_state.c_mu
    c_mu_prim = keps_state.c_mu_prim
    u = state.u
    v = state.v
    hz = state.grid.hz

    # prognostic computations
    _, bvf = compute_eos(state, case_tracable)
    shear2 = compute_shear(state, u, v)
    tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc = compute_tke_eps_bc(
        tke, hz, keps_params, case_tracable
    )

    # integrations
    tke_new = advance_turb(
        akt, akv, tke, tke, eps, c_mu, c_mu_prim, bvf, shear2, hz, dt,
        tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc, keps_params, True
    )
    eps_new = advance_turb(
        akt, akv, tke, tke_new, eps, c_mu, c_mu_prim, bvf, shear2, hz, dt,
        tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc, keps_params, False
    )

    # diagnostic variables
    akt_new, akv_new, eps_new, c_mu_new, c_mu_prim_new = compute_diag(
        tke_new, eps_new, bvf, shear2, keps_params
    )

    keps_state = eqx.tree_at(lambda t: t.akv, keps_state, akv_new)
    keps_state = eqx.tree_at(lambda t: t.akt, keps_state, akt_new)
    keps_state = eqx.tree_at(lambda t: t.tke, keps_state, tke_new)
    keps_state = eqx.tree_at(lambda t: t.eps, keps_state, eps_new)
    keps_state = eqx.tree_at(lambda t: t.c_mu, keps_state, c_mu_new)
    keps_state = eqx.tree_at(lambda t: t.c_mu_prim, keps_state, c_mu_prim_new)

    return keps_state

def compute_eos(state: State, case: CaseTracable) -> Tuple[ArrNzp1, ArrNz]:
    r"""
    Compute density anomaly and Brunt–Väisälä frequency.
    
    Prognostic computation via linear Equation Of State (EOS) :

    :math:`\rho = \rho_0(1-\alpha (T-T_0) + \beta (S-S_0))`

    :math:`N^2 = - \dfrac g {\rho_0} \partial_z \rho`

    Parameters
    ----------
    state : State
        Current state of the water column.
    case : CaseTracable
        Physical parameters and forcings of the model run.

    Returns
    -------
    bvf : float :class:`~jax.Array` of shape (nz+1)
        Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    rho : Float[Array, 'nz']
        Density anomaly :math:`\rho` on cell interfaces
        :math:`\left[\text {kg} \cdot \text m^{-3}\right]`

    Raises
    ------
    ValueError
        If the value of case.eos_tracers is not one of {'t', 's', 'ts', 'b'}.
    """
    rho0 = case.rho0
    match case.eos_tracers:
        case 't':
            rho = rho0 * (1. - case.alpha*(jnp.array(state.t)-case.t_rho_ref))
        case 's':
            rho = rho0 * (1. + case.beta*(jnp.array(state.s)-case.s_rho_ref))
        case 'ts':
            rho = rho0 * (1. - case.alpha*(jnp.array(state.t)-case.t_rho_ref) + \
                case.beta*(jnp.array(state.s)-case.s_rho_ref))
        case 'b':
            rho = rho0*(1-jnp.array(state.b)/case.grav)
        case _:
            raise ValueError("The attribute Case.eos_tracers must be one of {'t', 's', 'ts', 'b'}.")
    cff = 1./(state.grid.zr[1:]-state.grid.zr[:-1])
    bvf_in = - cff*case.grav/rho0 * (rho[1:]-rho[:-1])
    bvf = add_boundaries(0., bvf_in, float(bvf_in[-1]))
    return rho, bvf


def compute_shear(
        state: State,
        u_np1: ArrNz,
        v_np1: ArrNz
    ) -> ArrNzp1:
    r"""
    Compute shear production term for TKE equation.

    The prognostic equations are

    :math:`S_h^2 = \partial_Z U^n \cdot \partial_z U^{n+1/2}`

    where :math:`U^{n+1/2}` is the mean between :math:`U^n` and :math:`U^{n+1}`.
    
    Parameters
    ----------
    state : State
        Current state of the water column.
    u_np1 : float :class:`~jax.Array` of shape (nz)
        Zonal velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v_np1 : float :class:`~jax.Array` of shape (nz)
        Meridional velocity on the center of the cells at the next time step
        :math:`\left[\text m \cdot \text s^{-1}\right]`.            

    Returns
    -------
    shear2 : float :class:`~jax.Array` of shape (nz+1)
        Shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    """
    u_n = state.u
    v_n = state.v
    zr = state.grid.zr
    cff = 1.0 / (zr[1:] - zr[:-1])**2
    du = 0.5*cff * (u_np1[1:]-u_np1[:-1]) * (u_n[1:]+u_np1[1:]-u_n[:-1]-u_np1[:-1])
    dv = 0.5*cff * (v_np1[1:]-v_np1[:-1]) * (v_n[1:]+v_np1[1:]-v_n[:-1]-v_np1[:-1])
    shear2_in = du + dv
    return add_boundaries(0., shear2_in, 0.)


def compute_tke_eps_bc(
        tke: ArrNzp1,
        hz: ArrNz,
        keps_params: KepsParameters,
        case_tracable: CaseTracable
    ) -> Tuple[FloatJax, FloatJax, FloatJax, FloatJax]:
    r"""
    Compute top and bottom boundary conditions for TKE and :math:`\varepsilon`.

    Parameters
    ----------
    tke : float :class:`~jax.Array` of shape (nz+1)
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    hz : float :class:`~jax.Array` of shape (nz)
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.
    case_tracable : CaseTracable
        Physical parameters and forcings of the model run.

    Returns
    -------
    tke_sfc_bc : float
        TKE value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    tke_btm_bc : float
        TKE value for bottom boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    eps_sfc_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    eps_btm_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).

    Note
    ----
    The kind of boundary conditions between Neumann and Dirichlet are register in the parameters
    :code:`keps_params`.
    """
    # constants
    rp, rm, rn = keps_params.gls_p, keps_params.gls_m, keps_params.gls_n
    c_mu0 = keps_params.c_mu0
    cm0inv2 = 1./c_mu0**2
    vkarmn = case_tracable.vkarmn
    chk = keps_params.chk_grav/case_tracable.grav
    sig_eps = keps_params.sig_eps

    # velocity scales
    ustar2_sfc = jnp.sqrt(case_tracable.ustr_sfc**2 + case_tracable.vstr_sfc**2)
    ustar2_bot = jnp.sqrt(case_tracable.ustr_btm**2 + case_tracable.vstr_btm**2)

    # TKE Dirichlet boundary condition
    tke_sfc_dir = jnp.maximum(keps_params.tke_min, cm0inv2*ustar2_sfc)
    tke_btm_dir = jnp.maximum(keps_params.tke_min, cm0inv2*ustar2_bot)

    # TKE Neumann boundary condition
    tke_sfc_neu = 0.0
    tke_btm_neu = 0.0

    # epsilon surface conditions
    z0_s = jnp.maximum(keps_params.z0s_min, chk*ustar2_sfc)
    lgthsc = vkarmn*(0.5*hz[-1] + z0_s)
    tke_sfc = 0.5*(tke[-1]+tke[-2])
    eps_sfc_dir = jnp.maximum(keps_params.eps_min, c_mu0**rp * lgthsc**rn * tke_sfc**rm)
    eps_sfc_neu = float(-rn*vkarmn/sig_eps * c_mu0**(rp+1) * tke_sfc**(rm+.5) * lgthsc**rn)

    # epsilon bottom conditions
    z0b = jnp.maximum(keps_params.z0b, keps_params.z0b_min)
    lgthsc = vkarmn *(0.5*hz[0] + z0b)
    tke_btm = 0.5*(tke[0]+tke[1])
    eps_btm_dir = jnp.maximum(keps_params.eps_min, c_mu0**rp * lgthsc**rn * tke_btm**rm)
    eps_btm_neu = float(-rn*vkarmn/sig_eps * c_mu0**(rp+1) * tke_btm**(rm+.5) * lgthsc**rn)

    tke_sfc_bc = jnp.where(keps_params.dir_sfc, tke_sfc_dir, tke_sfc_neu)
    tke_btm_bc = jnp.where(keps_params.dir_btm, tke_btm_dir, tke_btm_neu)
    eps_sfc_bc = jnp.where(keps_params.dir_sfc, eps_sfc_dir, eps_sfc_neu)
    eps_btm_bc = jnp.where(keps_params.dir_btm, eps_btm_dir, eps_btm_neu)

    return tke_sfc_bc, tke_btm_bc, eps_sfc_bc, eps_btm_bc


def advance_turb(
        akt: ArrNzp1,
        akv: ArrNzp1,
        tke: ArrNzp1,
        tke_np1: ArrNzp1,
        eps: ArrNzp1,
        c_mu: ArrNzp1,
        c_mu_prim: ArrNzp1,
        bvf: ArrNzp1,
        shear2: ArrNzp1,
        hz: ArrNz,
        dt: float,
        tke_sfc_bc: FloatJax,
        tke_btm_bc: FloatJax,
        eps_sfc_bc: FloatJax,
        eps_btm_bc: FloatJax,
        keps_params: KepsParameters,
        do_tke: bool
    ) -> ArrNzp1:
    r"""
    Integrate TKE or :math:`\varepsilon` quantities.

    First the shear and buoyancy production are computed, then they are used in the building of the
    tridiagonal problem, the boundary conditions are then added and finally the tridiagonal problem
    is solved.

    Parameters
    ----------
    akt : float :class:`~jax.Array` of shape (nz+1)
        Current eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : float :class:`~jax.Array` of shape (nz+1)
        Current eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    tke : float :class:`~jax.Array` of shape (nz+1)
        Current turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    tke_np1 : float :class:`~jax.Array` of shape (nz+1)
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of the cells at next step
        (usefull only for :math:`\varepsilon` integration)
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps : float :class:`~jax.Array` of shape (nz+1)
        Current TKE dissipation denoted :math:`\varepsilon` on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu : float :class:`~jax.Array` of shape (nz+1)
        Current :math:`c_\mu` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    c_mu_prim : float :class:`~jax.Array` of shape (nz+1)
        Current :math:`c_\mu'` in GLS formalisim on the interfaces of the cells
        :math:`[\text{dimensionless}]`.
    bvf : float(nz+1)
        Current Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    shear2 : float :class:`~jax.Array` of shape (nz+1)
        Current shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    hz : float :class:`~jax.Array` of shape (nz)
        Thickness of cells from deepest to shallowest :math:`[\text m]`.
    dt : float
        Time-step of the forward model :math:`[\text s]`.
    tke_sfc_bc : float
        TKE value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    tke_btm_bc : float
        TKE value for bottom boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-3}\right]`).
    eps_sfc_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    eps_btm_bc : float
        :math:`\varepsilon` value for surface boundary condition (Dirichlet
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]` or Neumann
        :math:`\left[\text m ^3 \cdot \text s ^{-4}\right]`).
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.
    do_tke : bool
        If :code:`True` solve the equation for TKE, else for :math:`\varepsilon`.

    Returns
    -------
    vec : float :class:`~jax.Array` of shape (nz+1)
        TKE or :math:`\varepsilon` at next step (depending on :code:`do_tke`).
    """
    # fill the matrix off-diagonal terms for the tridiagonal problem
    cff = -0.5*dt
    ak_vec = jnp.where(do_tke, akv/keps_params.sig_k, akv/keps_params.sig_eps)
    a_in = cff*(ak_vec[1:-1]+ak_vec[:-2]) / hz[:-1]
    c_in = cff*(ak_vec[1:-1]+ak_vec[2:]) / hz[1:]

    # shear and buoyancy production
    s_prod_tke = akv[1:-1]*shear2[1:-1]
    b_prod_tke = -akt[1:-1]*bvf[1:-1]
    s_prod_eps = keps_params.c_eps1*c_mu[1:-1]*tke[1:-1]*shear2[1:-1]
    b_prod_eps = -c_mu_prim[1:-1] * tke[1:-1] * (keps_params.c_eps3m*jnp.maximum(bvf[1:-1], 0) + \
         keps_params.c_eps3p*jnp.minimum(bvf[1:-1], 0))
    s_prod = jnp.where(do_tke, s_prod_tke, s_prod_eps)
    b_prod = jnp.where(do_tke, b_prod_tke, b_prod_eps)

    # diagonal and f term
    cff = 0.5*(hz[:-1] + hz[1:])
    f_tke_in = lax.select(
        b_prod+s_prod > 0, cff*(tke[1:-1]+dt*(b_prod+s_prod)), cff*(tke[1:-1]+dt*s_prod)
    )
    f_eps_in = lax.select(
        b_prod+s_prod > 0, cff*(eps[1:-1]+dt*(b_prod+s_prod)), cff*(eps[1:-1]+dt*s_prod)
    )
    f_in = jnp.where(do_tke, f_tke_in, f_eps_in)
    b_tke_in = lax.select(
        (b_prod + s_prod) > 0, cff*(1. + dt*eps[1:-1]/tke[1:-1]) - a_in - c_in,
        cff*(1. + dt*(eps[1:-1] - b_prod)/tke[1:-1]) - a_in - c_in
    )
    b_eps_in = lax.select(
        (b_prod + s_prod) > 0,
        cff*(1. + dt*keps_params.c_eps2*eps[1:-1]/tke_np1[1:-1]) - a_in - c_in,
        cff*(1. + dt*keps_params.c_eps2*eps[1:-1]/tke_np1[1:-1] - dt*b_prod/eps[1:-1]) - a_in - c_in
    )
    b_in = jnp.where(do_tke, b_tke_in, b_eps_in)

    # surface boundary condition
    dir_sfc = keps_params.dir_sfc
    a_sfc = float(jnp.where(dir_sfc, 0., -0.5*(ak_vec[-1] + ak_vec[-2])))
    b_sfc = float(jnp.where(dir_sfc, 1., 0.5*(ak_vec[-1] + ak_vec[-2])))
    sfc_bc = jnp.where(do_tke, tke_sfc_bc, eps_sfc_bc)
    f_sfc = float(jnp.where(dir_sfc, sfc_bc, hz[-1]*sfc_bc))

    # bottom boundary condition
    dir_btm = keps_params.dir_btm
    b_btm = float(jnp.where(dir_sfc, 1., -0.5*(ak_vec[0] + ak_vec[1])))
    c_btm = float(jnp.where(dir_sfc, 0., 0.5*(ak_vec[0] + ak_vec[1])))
    btm_bc = jnp.where(do_tke, tke_btm_bc, eps_btm_bc)
    f_btm = float(jnp.where(dir_btm, btm_bc, hz[0]*btm_bc))

    # vectors rassembly
    a = add_boundaries(0., a_in, a_sfc)
    b = add_boundaries(b_btm, b_in, b_sfc)
    c = add_boundaries(c_btm, c_in, 0.)
    f = add_boundaries(f_btm, f_in, f_sfc)

    # solve tridiagonal problem
    f = tridiag_solve(a, b, c, f)

    vec_min = jnp.where(do_tke, keps_params.tke_min, keps_params.eps_min)
    vec = jnp.maximum(f, vec_min)

    return vec


def compute_diag(
        tke: ArrNzp1,
        eps: ArrNzp1,
        bvf: ArrNzp1,
        shear2: ArrNzp1,
        keps_params: KepsParameters
    ) -> Tuple[ArrNzp1, ArrNzp1, ArrNzp1, ArrNzp1, ArrNzp1]:
    r"""
    Computes the diagnostic variables of :math:`k-\varepsilon` closure.

    This function first apply the Galperin limitation, then it computes :math:`c_\mu'` and
    :math:`c_\mu` with the stability function, and finally it computes the eddy-diffusivity and
    viscosity with these variables.

    Parameters
    ----------
    tke : float :class:`~jax.Array` of shape (nz+1)
        Turbulent kinetic energy (TKE) denoted :math:`k` on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-2}\right]`.
    eps : float :class:`~jax.Array` of shape (nz+1)
        TKE dissipation denoted :math:`\varepsilon` on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    bvf : float :class:`~jax.Array` of shape (nz+1)
        Current Brunt–Väisälä frequency squared :math:`N^2` on cell interfaces
        :math:`\left[\text s^{-2}\right]`.
    shear2 : float :class:`~jax.Array` of shape (nz+1)
        Current shear production squared :math:`S_h^2` on cell interfaces
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    keps_params : KepsParameters
        Values of the parameters used by :math:`k-\varepsilon`.

    Returns
    -------
    akt : float :class:`~jax.Array` of shape (nz+1)
        Eddy-diffusivity on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    akv : float :class:`~jax.Array` of shape (nz+1)
        Eddy-viscosity on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    eps : float :class:`~jax.Array` of shape (nz+1)
        TKE dissipation denoted :math:`\varepsilon` on the interfaces of the cells at next step
        :math:`\left[\text m ^2 \cdot \text s ^{-3}\right]`.
    c_mu : float :class:`~jax.Array` of shape (nz+1)
        :math:`c_\mu` in GLS formalisim on the interfaces of the cells at next step
        :math:`[\text{dimensionless}]`.
    c_mu_prim : float :class:`~jax.Array` of shape (nz+1)
        :math:`c_\mu'` in GLS formalisim on the interfaces of the cells at next step
        :math:`[\text{dimensionless}]`.
    """
    # parameters
    akv_min = keps_params.akv_min
    akt_min = keps_params.akt_min
    eps_min = keps_params.eps_min
    sf_d0 = keps_params.sf_d0
    sf_d1 = keps_params.sf_d1
    sf_d2 = keps_params.sf_d2
    sf_d3 = keps_params.sf_d3
    sf_d4 = keps_params.sf_d4
    sf_d5 = keps_params.sf_d5
    sf_n0 = keps_params.sf_n0
    sf_n1 = keps_params.sf_n1
    sf_n2 = keps_params.sf_n2
    sf_nb0 = keps_params.sf_nb0
    sf_nb1 = keps_params.sf_nb1
    sf_nb2 = keps_params.sf_nb2
    lim_am0 = keps_params.lim_am0
    lim_am1 = keps_params.lim_am1
    lim_am2 = keps_params.lim_am2
    lim_am3 = keps_params.lim_am3
    lim_am4 = keps_params.lim_am4
    lim_am5 = keps_params.lim_am5
    lim_am6 = keps_params.lim_am6
    c_mu0 = keps_params.c_mu0
    rp, rm, rn = keps_params.gls_p, keps_params.gls_m, keps_params.gls_n
    e1 = 3 + rp/rn
    e2 = 1.5 + rm/rn
    e3 = -1/rn

    # minimum value of alpha_n to ensure that alpha_m is positive
    alpha_n_min = 0.5*(- (sf_d1 + sf_nb0) + jnp.sqrt((sf_d1 + sf_nb0)**2 - \
        4.0*sf_d0*(sf_d4 + sf_nb1))) / (sf_d4 + sf_nb1) # CHECK

    # Galperin limitation : l <= l_li
    l_lim = keps_params.galp*jnp.sqrt(2.0*tke[1:-1] / jnp.maximum(1e-14, bvf[1:-1]))

    # limitation (use MAX because rn is negative)
    cff = c_mu0**rp * l_lim**rn * tke[1:-1]**rm
    eps = eps.at[1:-1].set(jnp.maximum(eps[1:-1], cff))
    epsilon = c_mu0**e1 * tke[1:-1]**e2 * eps[1:-1]**e3
    epsilon = jnp.maximum(epsilon, eps_min)

    # compute alpha_n and alpha_m
    cff = (tke[1:-1] / epsilon)**2
    alpha_m = cff*shear2[1:-1]
    alpha_n = cff*bvf[1:-1]

    # limitation of alpha_n and alpha_m
    alpha_n = jnp.minimum(jnp.maximum(0.73*alpha_n_min, alpha_n), 1e10)
    alpha_m_max = (lim_am0 + lim_am1*alpha_n + lim_am2*alpha_n**2 + lim_am3*alpha_n**3) / \
        (lim_am4 + lim_am5*alpha_n + lim_am6*alpha_n**2)
    alpha_m = jnp.minimum(alpha_m, alpha_m_max)

    # compute stability functions
    denom = sf_d0 + sf_d1*alpha_n + sf_d2*alpha_m + sf_d3*alpha_n*alpha_m + sf_d4*alpha_n**2 + \
        sf_d5*alpha_m**2
    cff = 1./denom
    c_mu_in = cff*(sf_n0 + sf_n1*alpha_n + sf_n2*alpha_m)
    c_mu = add_boundaries(keps_params.c_mu_min, c_mu_in, keps_params.c_mu_min)
    c_mu_prim_in = cff*(sf_nb0 + sf_nb1*alpha_n + sf_nb2*alpha_m)
    c_mu_prim = add_boundaries(keps_params.c_mu_prim_min, c_mu_prim_in, keps_params.c_mu_prim_min)

    # finalize the computation of akv and akt
    cff = tke[1:-1]**2 / epsilon
    akt_in = jnp.maximum(cff*c_mu_prim[1:-1], akt_min)
    akv_in = jnp.maximum(cff*c_mu[1:-1], akv_min)

    akt_btm = float(jnp.maximum(1.5*akt_in[0] - 0.5*akt_in[1], akt_min))
    akt_sfc = float(jnp.maximum(1.5*akt_in[-1] - 0.5*akt_in[-2], akt_min))
    akv_btm = float(jnp.maximum(1.5*akv_in[0] - 0.5*akv_in[1], akv_min))
    akv_sfc = float(jnp.maximum(1.5*akv_in[-1] - 0.5*akv_in[-2], akv_min))

    akt = add_boundaries(akt_btm, akt_in, akt_sfc)
    akv = add_boundaries(akv_btm, akv_in, akv_sfc)

    return akt, akv, eps, c_mu, c_mu_prim



"""
Importation of Tunax classes and functions for shortcuts.
"""


from .space import Grid, State, Trajectory
from .case import Case, CaseTracable
from .closure import ClosureParametersAbstract, ClosureStateAbstract, Closure
from .database import Data, Obs, Database, Weights
from .functions import tridiag_solve, add_boundaries
from .closures_registry import CLOSURES_REGISTRY
from .model import (
    SingleColumnModel, lmd_swfrac, advance_tra_ed, advance_dyn_cor_ed, diffusion_solver
)
from .fitter import FittableParameter, FittableParametersSet



"""
Usefull calculation functions.

The functions in this module are supposed to be used in various other modules. They can be called by
the prefix :code:`tunax.functions.` or directly by :code:`tunax.`.

"""

from typing import Tuple, TypeAlias

import jax.numpy as jnp
from jax import lax
from jaxtyping import Float, Array

FloatJax: TypeAlias = Float[Array, '1']
"""Type that represent a float in a :class:`~jax.Array`, used only for the code linter."""


def tridiag_solve(
        a: Float[Array, 'n'],
        b: Float[Array, 'n'],
        c: Float[Array, 'n'],
        f: Float[Array, 'n']
    ) -> Float[Array, 'n']:
    r"""
    Solve a trigiagonal problem.

    The tridiagonal problem can be written :math:`\mathbb MX = F` where
    :math:`\mathbb M = \begin{pmatrix} b_1 & c_1 & & \\
    a_2 & \ddots & \ddots & \\
    & \ddots & \ddots & c_{n-1} \\
    & & a_n & b_n
    \end{pmatrix}`
    and :math:`F = \begin{pmatrix} f_1 \\ \vdots \\ f_n \end{pmatrix}`. The problem is solved by
    recurrence using :mod:`jax.lax.scan` 

    Parameters
    ----------
    a : float :class:`~jax.Array` of shape (n)
        Left diagonal of :math:`\mathbb M`, the first element is not used.
    b : float :class:`~jax.Array` of shape (n)
        Middle diagonal of :math:`\mathbb M`.
    c : float :class:`~jax.Array` of shape (n)
        Right diagonal of :math:`\mathbb M`, the last element is not used.
    f : float :class:`~jax.Array` of shape (n)
        Right hand of the equation :math:`F`.
    
    Returns
    -------
    x : float :class:`~jax.Array` of shape (n)
        Solution :math:`X` of tridiagonal problem.
    """
    def forward_scan_scal(
        carry: Tuple[FloatJax, FloatJax],
        x: Float[Array, '4']
    ) -> Tuple[Tuple[FloatJax, FloatJax], Tuple[FloatJax, FloatJax]]:
        f_im1, q_im1 = carry
        a, b, c, f = x
        cff = 1./(b+a*q_im1)
        f_i = cff*(f-a*f_im1)
        q_i = -cff*c
        carry = f_i, q_i
        return carry, carry
    init = f[0]/b[0], -c[0]/b[0]
    xs = jnp.stack([a, b, c, f])[:, 1:].T
    _, (f, q) = lax.scan(forward_scan_scal, init, xs)
    f = jnp.concat([jnp.array([init[0]]), f])
    q = jnp.concat([jnp.array([init[1]]), q])

    def reverse_scan_scal(carry: float, x: Float[Array, '2']) -> Tuple[float, float]:
        q_rev, f_rev = x
        carry = f_rev + q_rev*carry
        return carry, carry
    init = float(f[-1])
    xs = jnp.stack([q[::-1], f[::-1]])[:, 1:].T
    _, x = lax.scan(reverse_scan_scal, init, xs)
    x = jnp.concat([jnp.array([init]), x])

    return x[::-1]


def add_boundaries(
        vec_btm: float,
        vec_in: Float[Array, 'n-2'],
        vec_sfc: float
    ) -> Float[Array, 'n']:
    """
    Concatenate the three parts of a vector : surface, bottom and inside.

    This functions is made to avoid loops and make JAX more efficient by writing the calculations
    with vectorization when possible.

    Parameters
    ----------
    vec_btm : float
        Bottom value of the vector.
    vec_in : float :class:`~jax.Array` of shape (n-2)
        Middle values of the vector.
    vec_sfc : float
        Surface value of the vector.

    Returns
    -------
    vec : float :class:`~jax.Array` of shape (n)
        Concatenated vector.
    """
    return jnp.concat([jnp.array([vec_btm]), vec_in, jnp.array([vec_sfc])])


def _format_to_single_line(text: str) -> str:
    """
    Transforms a multiple line text in a line string by removing indentations.

    In the code the error and warning messages are written on multiple lines with indentations to
    respect the PEP8 maximum line length of 79 characters and the consistency of indentations. This
    function is used to show correctly these messages on one line and without the indentations.

    Parameters
    ----------
    text : str
        Text on multiple lines to transform.

    Returns
    -------
    line : str
        Text on a single line removed from indentations.
    """
    lines = text.splitlines()
    stripped_lines = [line.strip() for line in lines]
    single_line = " ".join(stripped_lines)
    return " ".join(single_line.split())



"""
Single column forward model core.

This module contains the main class :class:`SingleColumnModel` which is the core of Tunax for
implementing computing the evolution of a water column of the ocean. It also contains functions
that are used in this computation. The model was traduced from Frotran to JAX with the work of
Florian Lemarié and Manolis Perrot [1]_, the translation was done in part using the work of Anthony
Zhou, Linnia Hawkins and Pierre Gentine [2]_. this class and these functions can be obtained by the
prefix :code:`tunax.model.` or directly by :code:`tunax.`.

References
----------
.. [1] M. Perrot and F. Lemarié. Energetically consistent Eddy-Diffusivity Mass-Flux convective
    schemes. Part I: Theory and Models (2024). url :
    `hal.science/hal-04439113 <https://hal.science/hal-04439113>`_.
.. [2] A. Zhou, L. Hawkins and P. Gentine. Proof-of-concept: Using ChatGPT to Translate and
    Modernize an Earth System Model from Fortran to Python/JAX (2024). url :
    `arxiv.org/abs/2405.00018 <https://arxiv.org/abs/2405.00018>`_.

"""

from __future__ import annotations
import inspect
from typing import Tuple, Dict, TypeAlias

import equinox as eqx
import jax.numpy as jnp
from jax import lax, jit, vmap

from tunax.case import Case, CaseTracable
from tunax.space import Grid, State, Trajectory, ArrNz, ArrNzp1, ArrNt, VARIABLE_NAMES
from tunax.functions import FloatJax, tridiag_solve, add_boundaries
from tunax.closure import ClosureParametersAbstract, ClosureStateAbstract, Closure
from tunax.closures_registry import CLOSURES_REGISTRY

StatesTime: TypeAlias = Tuple[State, ClosureStateAbstract, float]
"""Type that represent the values that are transformed in an integration step of the model."""


class SingleColumnModel(eqx.Module):
    r"""
    Single column forward model core.

    This forward model of tunax is the combination of 4 things : the physical case
    :attr:`case_tracable`, an initial state of the water column :attr:`init_state`, the time
    information with :attr:`nt`, :attr:`dt` and :attr:`p_out` and the abstraction of the chosen
    closure for eddy-diffusivity :attr:`closure`. Adding a set of parameters for the closure one can
    run the model with the method :meth:`run`. The builder of this class takes the case as
    arguments all the attributes, except for the case : the parameter is an instance of
    :class:`~case.Case` which is transformed in a :class:`~case.CaseTracable` instance for JAX
    purpose and for the closure : the parameter is only the name of the closure.

    Parameters
    ----------
    nt : int
        cf. :attr:`nt`.
    dt : float
        cf. :attr:`dt`.
    p_out : int
        cf. :attr:`p_out`.
    init_state : State
        cf. :attr:`init_state`.
    case : Case
        Physical case and forcings of the experiment.
    closure_name : str
        Name of the chosen closure, must be a key of
        :attr:`~closures_registry.CLOSURES_REGISTRY`, see its documentation
        for the available closures.
    start_time : float, default=0.
        cf. :attr:`start_time`.

    Attributes
    ----------
    nt : int
        Number of integration interations.
    dt : float
        Time-step of integration for every iteration :math:`[\text s]`.
    p_out : int
        Number of of time-steps between every output step.
    init_state : State
        Initial physical state of the water column.
    case_tracable : CaseTracable
        Physical case and forcings of the experiment, made tracable for JAX purposes.
    closure : Closure
        Abstraction representing the chosen closure.
    start_time : float, default=0.
        Value of the starting time in :math:`[\text s]`.

    Note
    ----
    The closure parameters are given only at the end when we compute the model so that the
    calibrations of the parameters are easier.
        
    """

    nt: int = eqx.field(static=True)
    dt: float
    p_out: int = eqx.field(static=True)
    init_state: State
    case_tracable: CaseTracable
    closure: Closure = eqx.field(static=True)
    start_time: float = 0.

    def __init__(
            self,
            nt: int,
            dt: float,
            p_out: int,
            init_state: State,
            case: Case,
            closure_name: str,
            start_time: float=0.
        ) -> None:
        self.nt = nt
        self.dt = dt
        self.p_out = p_out
        self.init_state = init_state
        self.closure = CLOSURES_REGISTRY[closure_name]
        self.start_time = start_time
        # creation of the CaseTracable class
        grid = self.init_state.grid
        case_attributes = {k: v for k, v in vars(case).items() if not k.startswith('__')}
        for tra in ['t', 's', 'b', 'pt']:
            tra_attr = f'{tra}_forcing'
            tra_type_attr = f'{tra}_forcing_type'
            forcing = getattr(case, tra_attr)
            if forcing is not None:
                if isinstance(forcing, tuple):
                    case_attributes[tra_type_attr] = 'borders'
                    case_attributes[tra_attr] = forcing
                elif callable(forcing) and len(inspect.signature(forcing).parameters) == 1:
                    case_attributes[tra_type_attr] = 'constant'
                    vec_fun = vmap(forcing)
                    case_attributes[tra_attr] = grid.hz*vec_fun(grid.zr)
                elif callable(forcing) and len(inspect.signature(forcing).parameters) == 2:
                    case_attributes[tra_type_attr] = 'variable'
                    time = jnp.linspace(0, (self.nt-1)*self.dt, self.nt)
                    zr_grid, time_grid = jnp.meshgrid(grid.zr, time)
                    case_attributes[tra_attr] = grid.hz*forcing(zr_grid, time_grid)
            else:
                case_attributes[tra_type_attr] = None
        self.case_tracable = CaseTracable(**case_attributes)

    def tra_promote(self, promotions: Dict[str, str]) -> SingleColumnModel:
        """
        Increase the dimension type of the tracers.

        It is usefull to use :func:`~jax.vmap` on several instances of :class:`SingleColumnModel`
        to do batch computing (run the model in parallel).

        Parameters
        ----------
        promotions : Dict[str, str]
            The keys of the dictionnary are the name of the tracer variables to modify the
            dimensions of the forcing, one of {:code:`'t'`, :code:`'s'`, :code:`'b'`, :code:`'pt`'},
            the values are the new dimensions of the forcings, possible values
            {:code:`'constant'`, :code:`'variable'`}.
        
        Returns
        -------
        model : SingleColumnModel
            The :code:`self` object with the several promoted forcings.
        """
        case_tracable = self.case_tracable
        for tra, prom in promotions.items():
            grid = self.init_state.grid
            if prom == 'constant':
                case_tracable = case_tracable.tra_promote_borders_constant(tra, grid)
            elif prom == 'variable':
                initial_type = getattr(case_tracable, f'{tra}_forcing_type') == 'borders'
                if initial_type == 'borders':
                    case_tracable = case_tracable.tra_promote_borders_variable(tra, grid, self.nt)
                elif initial_type == 'borders':
                    case_tracable = case_tracable.tra_promote_constant_variable(tra, self.nt)
        return eqx.tree_at(lambda t: t.case_tracable, self, case_tracable)

    def step(
        self,
        state: State,
        closure_state: ClosureStateAbstract,
        time: float,
        closure_parameters: ClosureParametersAbstract
    ) -> StatesTime:
        r"""
        Runs one time-step of the model.
        

        This functions first call the closure to compute the eddy-diffusivity and viscosity, and
        then integrate the equations of tracers and momentum. It modifies the :code:`state` with
        these new values and then returns the new :code:`state` and :code:`closure_state`.
        
        Parameters
        ----------
        state : State
            State of the water column at the current :code:`time`.
        closure_state : ClosureStateAbstract
            State of the water column for the closure variables at the current :code:`time`.
        time : float
            Time of the current iteration (the mehtod integrates from this time to the next one).
        closure_parameters : ClosureParametersAbstract
            A set of parameters of the used closure.
        
        Returns
        -------
        state : State
            State of the water column after the integration.
        closure_state : ClosureStateAbstract
            State of the water column for the closure variables after the integration.
        time : float
            Value of the time after the integration.
        """
        dt = self.dt
        case_tracable = self.case_tracable
        # advance closure state (compute eddy-diffusivity and viscosity)
        closure_state = self.closure.step_fun(
            state, closure_state, dt, closure_parameters, case_tracable
        )
        # advance tracers
        i_time = int(time/self.dt)
        state = advance_tra_ed(state, closure_state.akt, dt, case_tracable, i_time)
        # advance velocities
        state = advance_dyn_cor_ed(state, closure_state.akv, dt, case_tracable)
        time += self.dt
        return state, closure_state, time

    def run_partial(
            self,
            state0: State,
            closure_state0: ClosureStateAbstract,
            time0: float,
            n_steps: int,
            closure_parameters: ClosureParametersAbstract
        ) ->  StatesTime:
        r"""
        Runs a certain number of time steps.

        Computes a loop of integration for a number of time steps of :code:`n_steps`, and return
        the last states of the loop.
        
        Parameters
        ----------
        state0 : State
            State of the water column at the beginning of the integration loop.
        closure_state : ClosureStateAbstract
            State of the water column for the closure variables at the beginning of the integration
            loop.
        time0 : float
            Begining time of the integration loop.
        n_steps : int
            Number of integration steps.
        closure_parameters : ClosureParametersAbstract
            A set of parameters of the used closure.    
        
        Returns
        -------
        state : State
            State of the water column after a number of :code:`n_steps` integration steps.
        closure_state : ClosureStateAbstract
            State of the water column for the closure variables after a number of :code:`n_steps`
            integration steps.
        time : float
            Value of the time after a number of :code:`n_steps` integration steps.
        """
        def scan_fn(carry: StatesTime, _: FloatJax) -> Tuple[StatesTime, None]:
            state, closure_state, time = carry
            state, closure_state, time = self.step(state, closure_state, time, closure_parameters)
            return (state, closure_state, time), None
        carry, _ = lax.scan(scan_fn, (state0, closure_state0, time0), jnp.arange(n_steps))
        (state, closure_state, time) = carry
        return state, closure_state, time

    def _state_concat_to_traj(self, states: State, times: ArrNt) -> Trajectory:
        """
        Convert the concatenations of :class:`~space.State` in a :class:`~space.Trajectory`.
        
        Use to get a trajectory from the output of the :func:`~jax.lax.scan` function which
        computes the concatenation of several :class:`~space.State` instances.

        Parameters
        ----------
        states : State
            A concatenation of :class:`~space.State` instances. More specifically, every leafs of
            the pytree is an array on the first axis of all the values of the the
            :class:`~space.State` instances.
        times : float :class:`~jax.Array` of shape (nt)
            Values of the different times corresponding at each :class:`~space.State` instances.
        
        Returns
        -------
        trajectory : Trajectory
            Trajectory corresponding to the concatenation of the states.
        """
        var_dict = {}
        for var in VARIABLE_NAMES:
            if getattr(states, var) is not None:
                var0 = jnp.expand_dims(getattr(self.init_state, var), 0)
                var_computed = getattr(states, var)
                var_dict[var] = jnp.concat([var0, var_computed])
        times = jnp.concat([jnp.array([self.start_time]), times])
        return Trajectory(self.init_state.grid, times, **var_dict)

    def run(self, closure_parameters: ClosureParametersAbstract) -> Trajectory:
        r"""
        Main run the model.

        Computes the :attr:`nt` integration steps of lenght :attr:`dt` from the initial state
        :attr:`init_state` with and doing an output of the state every :attr:`p_out` steps with the
        physical case and forcings corresponding to :attr:`case_tracable`. The closure for eddy-
        diffusivity used is :attr:`closure` with the values of the parameters described by the
        parameter :code:`closure_parameters`.
        
        Parameters
        ----------
        closure_parameters : ClosureParametersAbstract
            The set of parameters to use for the computation of the closure of eddy-diffusivity.
        
        Returns
        -------
        trajectory : Trajectory
            The trajectory with all the output steps of the integration loop.
        """
        init_closure_state = self.closure.state_class(self.init_state.grid, closure_parameters)

        def scan_fn(carry: StatesTime, _: FloatJax) -> Tuple[StatesTime, StatesTime]:
            state, closure_state, time = carry
            state, closure_state, time = self.run_partial(
                state, closure_state, time, self.p_out, closure_parameters
            )
            return (state, closure_state, time), (state, closure_state, time)

        n_steps_out = self.nt//self.p_out
        (_, _, _), (states, _, times) = lax.scan(
            scan_fn, (self.init_state, init_closure_state, self.start_time), jnp.arange(n_steps_out)
        )
        return self._state_concat_to_traj(states, jnp.array(times))

    @jit
    def jit_run(self, closure_parameters: ClosureParametersAbstract) -> Trajectory:
        r"""
        Jitted version of :meth:`run`.

        This method does exacly like :meth:`run` but :func:`~jax.jit` is applied on it, which
        means that the first call of this method will be the compilation of the function, and the
        next ones will be the compiled execution of the function which are supposed to be faster.
        There will be a compilation each time that this method will be call for a new "shape" of the
        :class:`SingleColumnModel` instance (which means that all the leafs of the pytree have the
        same shape as :class:`~jax.Array`), but even if this method is call on different instances,
        if they have the same shape, the compilation will be done only one time. Moreover, this
        method should use only for direct methods, is one wants to apply a :func:`~jax.grad` over
        the model, it's better to put the :func:`~jax.jit` outside.
        
        Parameters
        ----------
        closure_parameters : ClosureParametersAbstract
            cf. :meth:`run`.
        
        Returns
        -------
        trajectory : Trajectory
            cf. :meth:`run`.
        """
        return self.run(closure_parameters)


def lmd_swfrac(hz: ArrNz) -> ArrNzp1:
    r"""
    Compute solar forcing.

    Compute fraction of solar shortwave flux penetrating to specified depth due to exponential decay
    in Jerlov water type. This function is called once before running the model.

    Parameters
    ----------
    hz : float :class:`~jax.Array` of shape (nz)
        Thickness of cells from deepest to shallowest :math:`[\text m]`.

    Returns
    -------
    swr_frac : float :class:`~jax.Array` of shape (nz+1)
        Fraction of solar penetration throught the water column :math:`[\text{dimensionless}]`.
    """
    nz, = hz.shape
    mu1 = 0.35
    mu2 = 23.0
    r1 = 0.58
    attn1 = -1.0 / mu1
    attn2 = -1.0 / mu2

    xi1 = attn1 * hz
    xi2 = attn2 * hz

    def lax_step(sdwk, k):
        sdwk1, sdwk2 = sdwk
        sdwk1 = lax.cond(xi1[nz-k] > -20, lambda x: x*jnp.exp(xi1[nz-k]), lambda x: 0.*x, sdwk1)
        sdwk2 = lax.cond(xi2[nz-k] > -20, lambda x: x*jnp.exp(xi2[nz-k]), lambda x: 0.*x, sdwk2)
        return (sdwk1, sdwk2), sdwk1+sdwk2

    _, swr_frac = lax.scan(lax_step, (r1, 1.0 - r1), jnp.arange(1, nz+1))
    return jnp.concat((swr_frac[::-1], jnp.array([1.])))


def tracer_flux(
        tracer: str,
        case_tracable: CaseTracable,
        grid: Grid,
        i_time: int
    ) -> ArrNz:
    r"""
    Computes flux of the tracer forcing.
    
    This function get the flux of the forcing at a certain time depending on the type of the
    forcing, the flux being the derivative of the forcing along the depth.
    
    Parameters
    ----------
    tracer : str
        Name of the tracer variable of the concerned forcing. One of {:code:`'t'`, :code:`'s'`,
        :code:`'b'`, :code:`'pt`'}.
    case_tracable : CaseTracable
        Physical case which contains the forcings type and values.
    grid : Grid
        Vertical grid of the water column.
    i_time : int
        Index of the time iteration corresponding to the index in the forcing.
    
    Returns
    -------
    df : float :class:`~jax.Array` of shape (nz)
        Flux of the forcing of the tracer. At each cell it represents the difference between the
        input and the ouput flux.
    
    Raises
    ------
    ValueError
        If the forcing type is not one of {'borders', 'constant', 'variable'}.
    """
    forcing = getattr(case_tracable, f'{tracer}_forcing')
    forcing_type = getattr(case_tracable, f'{tracer}_forcing_type')
    match forcing_type:
        case 'borders':
            df = add_boundaries(-forcing[0], jnp.zeros(grid.nz-2), forcing[1])
        case 'constant':
            df = forcing
        case 'variable':
            df = forcing[:, i_time]
        case _:
            mess = f'Forcing type of variable {tracer} should be one of' + \
                "{'borders', 'constant', 'variable'}."
            raise ValueError(mess)
    return df


def advance_tra_ed(
        state: State,
        akt: ArrNzp1,
        dt: float,
        case_tracable: CaseTracable,
        i_time: int,
    ) -> State:
    r"""
    Integrate vertical diffusion term for tracers.

    First the flux divergences are computed taking in account the forcings. Then the diffusion
    equation of the tracers system is solved, and the tracers at next time-step are returned. The
    solved equation is for :math:`C` a tracer :

    :math:`\partial _z ( K_m \partial _z C) + \partial _t C + F = 0`

    where :math:`F` is the representation of the forcings. This equation is solved for every tracer
    indicated in :attr:`~case.Case.eos_tracers` and the passive tracer if :attr:`~case.Case.do_pt`
    is set.
    
    Parameters
    ----------
    state : State
        State of the water column at the current iteration.
    akt : float :class:`~jax.Array` of shape (nz)+1
        Eddy-diffusivity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    dt : float
        Time-step of the integration step.
    case_tracable : CaseTracable
        Physical case which contains the forcings type and values.
    i_time : int
        Index of the time iteration corresponding to the index in the forcing.
    
    Returns
    -------
    state : State
        The state of the water column with the values of the tracers after the integration and the
        diffusion by the eddy-diffusivity.
    """
    hz = state.grid.hz
    tracers = [tra for tra in case_tracable.eos_tracers]
    if case_tracable.do_pt:
        tracers.append('pt')
    def get_pytree_fun(tracer: str):
        return lambda t: getattr(t, tracer)
    for tracer in tracers:
        tra = getattr(state, tracer)
        df = tracer_flux(tracer, case_tracable, state.grid, i_time)
        df = hz*tra + dt*df
        tra = diffusion_solver(akt, hz, df, dt)
        state = eqx.tree_at(get_pytree_fun(tracer), state, tra)
    return state


def advance_dyn_cor_ed(
        state: State,
        akv: ArrNzp1,
        dt: float,
        case_tracable: CaseTracable
    ) -> State:
    r"""
    Integrate vertical diffusion and Coriolis terms for momentum.

    First the Coriolis term is computed, then the momentum forcings are applied and finally, the
    diffusion equation is solved. The momentum at next time-step is returned. The equation which is
    solved is :

    :math:`\partial_z (K_v \partial_z U) + F_{\text{cor}}(U) + F = 0`

    where :math:`F_{\text{cor}}` represent the Coriolis effect, and :math:`F` represent the effect
    of the forcings on the momentum.
    
    Parameters
    ----------
    state : State
        State of the water column at the current iteration.
    akv : float :class:`~jax.Array` of shape (nz+1)
        Eddy-viscosity on the interfaces of the cells
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    dt : float
        Time-step of the integration step.
    case_tracable : CaseTracable
        Physical case which contains the forcings type and values.
    
    Returns
    -------
    state : State
        The state of the water column with the values of the momentum after the integration and the
        diffusion by the eddy-viscosity.
    """
    gamma_cor = 0.55
    fcor = case_tracable.fcor
    u = state.u
    v = state.v
    hz = state.grid.hz

    # 1 - Compute Coriolis term
    cff = (dt * fcor) ** 2
    cff1 = 1 / (1 + gamma_cor * gamma_cor * cff)
    fu = cff1 * hz * ((1-gamma_cor*(1-gamma_cor)*cff)*u + dt*fcor*v)
    fv = cff1 * hz * ((1-gamma_cor*(1-gamma_cor)*cff)*v - dt*fcor*u)

    # 2 - Apply surface and bottom forcing
    fu = fu.at[-1].add(dt * case_tracable.ustr_sfc)
    fv = fv.at[-1].add(dt * case_tracable.vstr_sfc)
    fu = fu.at[0].add(-dt * case_tracable.ustr_btm)
    fv = fv.at[0].add(-dt * case_tracable.vstr_btm)

    # 3 - Implicit integration for vertical viscosity
    u = diffusion_solver(akv, hz, fu, dt)
    v = diffusion_solver(akv, hz, fv, dt)

    # 4 - Update the state
    state = eqx.tree_at(lambda t: t.u, state, u)
    state = eqx.tree_at(lambda t: t.v, state, v)

    return state


def diffusion_solver(
        ak: ArrNzp1,
        hz: ArrNz,
        f: ArrNz,
        dt: float
    ) -> ArrNz:
    r"""
    Solve a diffusion problem with finite volumes.

    The diffusion problems can be written

    :math:`\partial _z (K \partial _z X) + \dfrac f {\Delta t \Delta x} = 0`

    where we are searching for :math:`X` and where :math:`f` represents the temporal derivative and
    forcings. This function transforms this problem in a tridiagonal system and then solve it.

    Parameters
    ----------
    ak : float :class:`~jax.Array` of shape (nz+1)
        Diffusion at the cell interfaces :math:`K` in
        :math:`\left[\text m ^2 \cdot \text s ^{-1}\right]`.
    hz : float :class:`~jax.Array` of shape (nz)
        Thickness of cells from deepest to shallowest :math:`\left[\text m\right]`.
    f : float :class:`~jax.Array` of shape (nz)
        Right-hand flux of the equation :math:`f` in :math:`[[X] \cdot \text m ]`.
    dt : float
        Time-step of discretisation :math:`[\text s]`.
    
    Returns
    -------
    x : float :class:`~jax.Array` of shape (nz)
        Solution of the diffusion problem :math:`X` in :math:`\left[[X]\right]`.
    """
    # fill the coefficients for the tridiagonal matrix
    a_in = -2.0 * dt * ak[1:-2] / (hz[:-2] + hz[1:-1])
    c_in = -2.0 * dt * ak[2:-1] / (hz[2:] + hz[1:-1])
    b_in = hz[1:-1] - a_in - c_in

    # bottom boundary condition
    c_btm = -2.0 * dt * ak[1] / (hz[1] + hz[0])
    b_btm = hz[0] - c_btm

    # surface boundary condition
    a_sfc = -2.0 * dt * ak[-2] / (hz[-2] + hz[-1])
    b_sfc = hz[-1] - a_sfc

    # concatenations
    a = add_boundaries(0., a_in, float(a_sfc))
    b = add_boundaries(float(b_btm), b_in, float(b_sfc))
    c = add_boundaries(float(c_btm), c_in, 0.)

    x = tridiag_solve(a, b, c, f)

    return x



"""
Physical parameters and forcings.

This module contains the :class:`Case` class which is the one used to describe the parameters and
the forcings of a model. It also contains the :class:`CaseTracable` which is a variation of the
first classe which works better with JAX specificities. This classes can be obtained by the prefix
:code:`tunax.case.` or directly by :code:`tunax.`.

"""

from __future__ import annotations
from typing import Union, Tuple, Callable, Optional, TypeAlias
from dataclasses import replace

import equinox as eqx
import jax.numpy as jnp
from jax import device_get

from tunax.functions import add_boundaries
from tunax.space import Grid, ArrNz, ArrNzNt

ForcingType: TypeAlias = Union[
    Tuple[float, float],
    Callable[[float], float],
    Callable[[float, float], float]
]
"""Type that represent the different possible types of the forcings in :class:`Case`."""
ForcingArrayType: TypeAlias = Union[Tuple[float, float], ArrNz, ArrNzNt]
"""Type that represent the different possible types of the forcings in :class:`CaseTracable`."""

_OMEGA = 7.292116e-05
"""Rotation rate of the Earth [rad.s-1]."""
_RAD_DEG = jnp.pi/180.
"""Measure of one degree in radiant [rad.°-1]."""


class Case(eqx.Module):
    r"""
    Physical parameters and forcings.

    This class contains all the physical constants, and the constant forcings that definine an
    experience for the model. The forcings can be described as functions. The transformations from
    :class:`Case` to :class:`CaseTracable` is done when a model instance is created. The constructor
    takes all the attributes as parameters.

    Attributes
    ----------
    rho0 : float, default=1024.
        Default density of saltwater :math:`[\text{kg} \cdot \text{m}^{-3}]`.
    grav : float, default=9.81
        Gravity acceleration :math:`[\text{m} \cdot \text{s}^{-2}]`.
    cp : float, default=3985.
        Specific heat capacity of saltwater
        :math:`[\text{J} \cdot \text{kg}^{-1} \cdot \text{K}^{-1}]`.
    eos_tracers : str, default='t'
        Tracers used for the equation of state (the computation of the density). One of
        {:code:`'t'`, :code:`'s'`, :code:`'ts`', :code:`'b'`}.
        - 't': temperature
        - 's': salinity
        - 'ts': temperature and salinity
        - 'b': buoyancy
    alpha : float, default=2e-4
        Thermal expansion coefficient :math:`[\text{K}^{-1}]`.
    beta : float, default=8e-4
        Salinity expansion coefficient :math:`[\text{psu}^{-1}]`.
    t_rho_ref : float, default=0.
        Reference temperature for the equation of state (the computation of the density)
        :math:`[° \text C]`.
    s_rho_ref : float, default=35.
        Reference salinity for the equation of state (the computation of the density)
        :math:`[\text{psu}]`.
    do_pt : bool, default=False
        Compute or not a passive tracer.
    vkarmn : float, default=0.384
        Von Kármán constant :math:`[\text{dimensionless}]`.
    fcor : float, default=0.
        Coriolis frequency at the water column :math:`[\text{rad} \cdot \text{s}^{-1}]`.
    ustr_sfc : float, default=0.
        Zonal wind stress :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    ustr_btm : float, default=0.
        Zonal current stress at the bottom :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    vstr_sfc : float, default=0.
        Meridional wind stress :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    vstr_btm : float, default=0.
        Meridional current stress at the bottom :math:`[\text{m}^{2} \cdot \text{s}^{-2}]`.
    t_forcing : tuple of 2 floats or a function, optionnal, default=None
        Description of the forcing of temperature (potentially no forcing if the variable is not
        activated i.e. if :code:`'t'` is not in :code:`eos_tracers`). There are 3 cases :

        - **Border forcing** : tuple of 2 floats, the first one is the forcing at the bottom and the
          second ont is the forcing at the top of the water column, the unit is in
          :math:`[\text{K} \cdot \text{m} \cdot \text{s}^{-1}]`.
        
        - **Deep constant forcing** : function of signature float->float, the parameter is the depth
          and the ouput is the value of the forcing at this depth in
          :math:`[\text{K} \cdot \text{s}^{-1}]`. The values of the functions represent the flux of
          the forcing (the derivative along the depth).)

        - **Deep variable forcing** : function of signature (float, float)->float, the parameters
          are the depth and the time and the ouput is the value of the forcing at this depth and
          this time in :math:`[\text{K} \cdot \text{s}^{-1}]`. The values of the functions represent
          the flux of the forcing (the derivative along the depth).

    s_forcing : tuple of 2 floats or a function, optionnal, default=None
        Description of the forcing of salinity (potentially no forcing if the variable is not
        activated i.e. if :code:`'s'` is not in :code:`eos_tracers`). There are the 3 same cases as
        for the :code:`t_forcing`, and the units are
        :math:`[\text{psu} \cdot \text{m} \cdot \text{s}^{-1}]` for the border forcing and
        :math:`[\text{psu} \cdot \text{s}^{-1}]` for the other ones.
    b_forcing : tuple of 2 floats or a function, optionnal, default=None
        Description of the forcing of buoyancy (potentially no forcing if the variable is not
        activated i.e. if :code:`'b'` is not in :code:`eos_tracers`). There are the 3 same cases as
        for the :code:`t_forcing`, and the units are
        :math:`[\text{m} \cdot \text{s}^{-1}]` for the border forcing and
        :math:`[\text{s}^{-1}]` for the other ones.
    pt_forcing : tuple of 2 floats or a function, optionnal, default=None
        Description of the forcing of passive tracer (potentially no forcing if the variable is not
        activated i.e. if :code:`do_pt` is not set). There are the 3 same cases as for the
        :code:`t_forcing`, and the units are :math:`[\text{m} \cdot \text{s}^{-1}]`
        for the border forcing and :math:`[\text{s}^{-1}]` for the other ones.

    """

    # physcal constants
    rho0: float = 1024.
    grav: float = 9.81
    cp: float = 3985.
    eos_tracers: str = eqx.field(default='t', static=True)
    alpha: float = 2e-4
    beta: float = 8e-4
    t_rho_ref: float = 0.
    s_rho_ref: float = 35.
    do_pt: bool = eqx.field(default=False, static=True)
    vkarmn: float = 0.384
    # dynamic forcings
    fcor: float = 0.
    ustr_sfc: float = 0.
    ustr_btm: float = 0.
    vstr_sfc: float = 0.
    vstr_btm: float = 0.
    # tracers forcings
    t_forcing: Optional[ForcingType] = eqx.field(default=None, static=True)
    s_forcing: Optional[ForcingType] = eqx.field(default=None, static=True)
    b_forcing: Optional[ForcingType] = eqx.field(default=None, static=True)
    pt_forcing: Optional[ForcingType] = eqx.field(default=None, static=True)

    def set_lat(self, lat: float) -> Case:
        """
        Set the Coriolis frequency from the latitude.

        Parameters
        ----------
        lat : float
            Latitude of the water column :math:`[°]`.
        
        Returns
        -------
        case : Case
            The :code:`self` object with the the new value of :attr:`fcor`.
        """
        fcor = float(device_get(2.*_OMEGA*jnp.sin(_RAD_DEG*lat)))
        case = eqx.tree_at(lambda t: t.fcor, self, fcor)
        return case


class CaseTracable(eqx.Module):
    r"""
    Physical parameters and forcings tracable by JAX.

    This class is similar to the :class:`Case` one, but the function forcings are transformed in
    arrays so that the class is tracable by JAX which means that we can use :func:`~jax.jit` and
    :func:`~jax.grad` more freely. The constructor takes all the attributes as parameters.

    Attributes
    ----------
    rho0 : float, default=1024.
        cf. :attr:`Case.rho0`
    grav : float, default=9.81
        cf. :attr:`Case.grav`
    cp : float, default=3985.
        cf. :attr:`Case.cp`
    eos_tracers : str, default='t'
        cf. :attr:`Case.eos_tracers`
    alpha : float, default=2e-4
        cf. :attr:`Case.alpha`
    beta : float, default=8e-4
        cf. :attr:`Case.beta`
    t_rho_ref : float, default=0.
        cf. :attr:`Case.t_rho_ref`
    s_rho_ref : float, default=35.
        cf. :attr:`Case.s_rho_ref`
    do_pt : bool, default=False
        cf. :attr:`Case.do_pt`
    vkarmn : float, default=0.384
        cf. :attr:`Case.vkarmn`
    fcor : float, default=0.
        cf. :attr:`Case.fcor`
    ustr_sfc : float, default=0.
        cf. :attr:`Case.ustr_sfc`
    ustr_btm : float, default=0.
        cf. :attr:`Case.ustr_btm`
    vstr_sfc : float, default=0.
        cf. :attr:`Case.vstr_sfc`
    vstr_btm : float, default=0.
        cf. :attr:`Case.vstr_btm`
    t_forcing : tuple of 2 floats or :class:`~jax.Array` (nz) or (nz, nt), optionnal, default=None
        Description of the temperature forcing cf. :attr:`Case.t_forcing`, the type depends on the
        forcing type :
        
        - **Border forcing** : tuple of 2 floats, the first one is the forcing at the bottom and the
          second one is the forcing at the top of the water column, the unit is in
          :math:`[\text{K} \cdot \text{m} \cdot \text{s}^{-1}]`.
        
        - **Deep constant forcing** : array of shape (nz) : the value of the forcing function on the
          geometrical :class:`Grid` of the model. The values represent the forcing flux, which is
          for each cell the difference between the forcing at the top of the cell and the forcing at
          bottom.
        
        - **Deep variable forcing** : array of shape (nz, nt) : the value of the forcing function on
          the geometrical :class:`Grid` and the different iteration times of the model. As for deep
          constant forcing, the values represent the flux of the forcing at every time.

    s_forcing : tuple of 2 floats or :class:`~jax.Array` (nz) or (nz, nt), optionnal, default=None
        Same as :attr:`t_forcing` for Salinity.
    b_forcing : tuple of 2 floats or :class:`~jax.Array` (nz) or (nz, nt), optionnal, default=None
        Same as :attr:`t_forcing` for buoyancy.
    pt_forcing : tuple of 2 floats or :class:`~jax.Array` (nz) or (nz, nt), optionnal, default=None
        Same as :attr:`t_forcing` for passive tracer.
    t_forcing_type : str, optionnal, default=None
        Description of the type of temperature forcing : :code:`'borders'` for **border forcing**,
        :code:`'constant'` for **deep constant forcing** and :code:`'variable'` for **deep variable
        forcing**
    s_forcing_type : str, optionnal, default=None
        Same as :attr:`t_forcing_type` for salinity.
    b_forcing_type : str, optionnal, default=None
        Same as :attr:`t_forcing_type` for buoyancy.
    pt_forcing_type : str, optionnal, default=None
        Same as :attr:`t_forcing_type` for passive tracer.

    """

    # physcal constants
    rho0: float = 1024.
    grav: float = 9.81
    cp: float = 3985.
    eos_tracers: str = eqx.field(default='t', static=True)
    alpha: float = 2e-4
    beta: float = 8e-4
    t_rho_ref: float = 0.
    s_rho_ref: float = 35.
    do_pt: bool = eqx.field(default=False, static=True)
    vkarmn: float = 0.384
    # dynamic forcings
    fcor: float = 0.
    ustr_sfc: float = 0.
    ustr_btm: float = 0.
    vstr_sfc: float = 0.
    vstr_btm: float = 0.
    # tracers forcings
    t_forcing: Optional[ForcingArrayType] = None
    s_forcing: Optional[ForcingArrayType] = None
    b_forcing: Optional[ForcingArrayType] = None
    pt_forcing: Optional[ForcingArrayType] = None
    t_forcing_type: Optional[str] = eqx.field(default=None, static=True)
    s_forcing_type: Optional[str] = eqx.field(default=None, static=True)
    b_forcing_type: Optional[str] = eqx.field(default=None, static=True)
    pt_forcing_type: Optional[str] = eqx.field(default=None, static=True)

    def tra_promote_borders_constant(self, tra: str, grid: Grid) -> CaseTracable:
        """
        Promote the dimension of a forcing from borders to constant.

        It can be use to apply :func:`~jax.vmap` on :class:`SingleColumnModel` for batch computing.
        The input :class:`CaseTracable` instance should have a borders forcing type and the ouput
        instance will have deep constant forcing type.
        
        Parameters
        ----------
        tra : str
            Name of the tracer variable of the concerned forcing. One of {:code:`'t'`, :code:`'s'`,
            :code:`'b`', :code:`'pt`'}.
        grid : Grid
            Spatial grid to compute the constant forcing on.
        
        Returns
        -------
        case_tracable : CaseTracable
            The :code:`self` object with the promoted forcing.
        """
        border_forcing = getattr(self, f'{tra}_forcing')
        constant_forcing = add_boundaries(
            -border_forcing[0], jnp.zeros(grid.nz-2), border_forcing[1]
        )
        case_tracable = eqx.tree_at(lambda t: getattr(t, f'{tra}_forcing'), self, constant_forcing)
        dico = {f'{tra}_forcing_type': 'constant'}
        case_tracable = replace(case_tracable, **dico)
        return case_tracable

    def tra_promote_borders_variable(self, tra: str, grid: Grid, nt: int) -> CaseTracable:
        """
        Promote the dimension of a forcing from borders to variable.

        It can be use to apply :func:`~jax.vmap` on :class:`SingleColumnModel` for batch computing.
        The input :class:`CaseTracable` instance should have a borders forcing type and the ouput
        instance will have deep variable forcing type.
        
        Parameters
        ----------
        tra : str
            Name of the tracer variable of the concerned forcing. One of {:code:`'t'`, :code:`'s'`,
            :code:`'b`', :code:`'pt`'}.
        grid : Grid
            Spatial grid to compute the constant forcing on.
        nt : int
            Number of integration iterations of the model.
        
        Returns
        -------
        case_tracable : CaseTracable
            The :code:`self` object with the promoted forcing.
        """
        border_forcing = getattr(self, f'{tra}_forcing')
        constant_forcing = add_boundaries(
            -border_forcing[0], jnp.zeros(grid.nz-2), border_forcing[1]
        )
        variable_forcing = jnp.tile(constant_forcing[:, None], (1, nt))
        case_tracable = eqx.tree_at(lambda t: getattr(t, f'{tra}_forcing'), self, variable_forcing)
        dico = {f'{tra}_forcing_type': 'variable'}
        case_tracable = replace(case_tracable, **dico)
        return self

    def tra_promote_constant_variable(self, tra: str, nt: int) -> CaseTracable:
        """
        Promote the dimension of a forcing from borders to constant.

        It can be use to apply :func:`~jax.vmap` on :class:`~model.SingleColumnModel` for batch
        computing. The input :class:`CaseTracable` instance should have a constant forcing type and
        the ouput instance will have deep variable forcing type.
        
        Parameters
        ----------
        tra : str
            Name of the tracer variable of the concerned forcing. One of {:code:`'t'`, :code:`'s'`,
            :code:`'b'`, :code:`'pt`'}.
        grid : Grid
            Spatial grid to compute the constant forcing on.
        nt : int
            Number of integration iterations of the model.
        
        Returns
        -------
        case_tracable : CaseTracable
            The :code:`self` object with the promoted forcing.
        """
        constant_forcing = getattr(self, f'{tra}_forcing')
        variable_forcing = jnp.tile(constant_forcing[:, None], (1, nt))
        case_tracable = eqx.tree_at(lambda t: getattr(t, f'{tra}_forcing'), self, variable_forcing)
        dico = {f'{tra}_forcing_type': 'variable'}
        case_tracable = replace(case_tracable, **dico)
        return self



"""
Geometry and variables of the model.

This module contains the objects that are used in Tunax to describe the geometry of the water column
in :class:`Grid`, the variables of the water column at one time-step in :class:`State` and the time-
series of the model computation in :class:`Trajectory`. These classes can be obtained by the
prefix :code:`tunax.space.` or directly by :code:`tunax.`.

"""

from __future__ import annotations
from typing import Optional, List, Dict, Callable, TypeAlias, cast
import warnings

import equinox as eqx
import xarray as xr
import jax.numpy as jnp
from jax import vmap
from jaxtyping import Float, Array

from tunax.functions import _format_to_single_line

ArrNz: TypeAlias = Float[Array, 'nz']
"""Type that describes a float :class:`~jax.Array` of shape (nz)."""
ArrNzp1: TypeAlias = Float[Array, 'nz+1']
"""Type that describes a float :class:`~jax.Array` of shape (nz+1)."""
ArrNt: TypeAlias = Float[Array, 'nt']
"""Type that describes a float :class:`~jax.Array` of shape (nt)."""
ArrNzNt: TypeAlias = Float[Array, 'nz nt']
"""Type that describes a float :class:`~jax.Array` of shape (nz, nt)."""

TRACERS_NAMES: List[str] = ['t', 's', 'b', 'pt']
"""Names of the tracers, in the order of temperature, salinity, buoyancy and passive tracer."""
VARIABLE_NAMES: List[str] = ['u', 'v'] + TRACERS_NAMES
"""Names of all the variables, zonal and meridionnal velocities in addition of the tracers."""
VARIABLE_SHAPES: Dict[str, str] = {
    'u': 'zr',
    'v': 'zr',
    't': 'zr',
    's': 'zr',
    'b': 'zr',
    'pt': 'zr'
}
"""Shapes of all the variables on the water column."""


def _piecewise_linear_ramp(z: float, z0: float, f0: float)-> float:
    r"""
    Mathemacial function used for the state initialisation.

    Apply to z a function linear by part and continuous :
    f(z) = 0 if z < zm and f(z) = f0 (1-z/zm) else.

    Parameters
    ----------
    z : float
        The value where to apply the function.
    zm : float
        The point of connexion of the two linear parts of the function.
    f0 : float
        The value of the function in 0.

    Returns
    -------
    fz : float
        The value of the function in z.
    """
    return f0*(z/-z0+1) * (z>z0)

def _piecewise_linear_flat(z: float, zm: float, f0: float, sl: float) -> float:
    r"""
    Mathemacial function used for the state initialisation.

    Apply to z a function linear by part and continuous :
    f(z) = f0 + s_l (z-zm) if z < z_m and f(z) = f0 else.

    Parameters
    ----------
    z : float
        The value where to apply the function.
    zm : float
        The point of connexion of the two linear parts of the function.
    f0 : float
        The value of the function in 0 and in the right part of the
        funcion.
    sl : float
        The slope of the left part of the function.

    Returns
    -------
    fz : float
        The value of the function in z.
    """
    return f0 + sl*(z-zm) * (z<zm)


class Grid(eqx.Module):
    r"""
    One dimensional spatial geometry of a water column.

    This mesh is made up of a number of :attr:`nz` of cells (:attr:`zr`) of potentially varying
    thickness (:attr:`hz`), separated by interface points (:attr:`zw`) and extending from the ocean
    surface at a depth of :math:`0` to the ocean floor at a depth of :attr:`hbot`.

    Parameters
    ----------
    zr : float :class:`~jax.Array` of shape (nz)
        cf. :attr:`zr`.
    zw : float :class:`~jax.Array` of shape (nz+1)
        cf. :attr:`zw`.

    Attributes
    ----------
    nz : int
        Number of cells.
    hbot : float
        Depth of the water column :math:`[\text m]`.
    zr : float :class:`~jax.Array` of shape (nz)
        Depths of cell centers from deepest to shallowest :math:`[\text m]`
    zw : float :class:`~jax.Array` of shape (nz+1)
        Depths of cell interfaces from deepest to shallowest :math:`[\text m]`.
    hz : float :class:`~jax.Array` of shape (nz)
        Thickness of cells from deepest to shallowest :math:`[\text m]`.

    Note
    ----
    The constructor :code:`__init__` takes only :attr:`zr` and :attr:`zw` as as arguments and
    construct the other attributes from them. The centers of the cells :attr:`zr` are not necessarly
    the middle between the interfaces :attr:`zw` but should be between.

    """

    nz: int = eqx.field(static=True)
    hbot: float
    zr: ArrNz
    zw: ArrNzp1
    hz: ArrNz

    def __init__(self, zr: ArrNz, zw: ArrNzp1) -> None:
        self.nz = zr.shape[0]
        self.hbot = float(zw[0])
        self.zw = zw
        self.zr = zr
        self.hz = zw[1:] - zw[:-1]

    def find_index(self, h: float) -> int:
        r"""
        Find the index of a depth.

        Find the index :code:`i` so that the depth :code:`h` is in cell :code:`i`, which means
        :math:`z^w_i \leqslant -h \leqslant z^w_{i+1}` if :math:`h \leqslant 0` and :math:`i=-1` if
        :math:`h>0`.

        Parameters
        ----------
        h : float, positive
            The depth to search the index :math:`[\text m]`.

        Returns
        -------
        i : int
            The index corresponding to the depth :code:`h`.
        """
        return int(jnp.searchsorted(self.zw, -h, side='right')) - 1

    @classmethod
    def linear(cls, nz: int, hbot: int) -> Grid:
        r"""
        Creates a grid with equal thickness cells.

        The grid instance will have :attr:`nz` cells of equal thickness for a depth of :attr:`hbot`.

        Parameters
        ----------
        nz : int
            Number of cells.
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.

        Returns
        -------
        grid : Grid
            The linear grid.
        """
        zw = jnp.linspace(-hbot, 0, nz+1)
        zr = 0.5*(zw[:-1]+zw[1:])
        return cls(zr, zw)

    @classmethod
    def analytic(cls, nz: int, hbot: float, hc: float, theta: float=6.5) -> Grid:
        r"""
        Creates a grid of type analytic.

        The grid instance will have a depth of :attr:`hbot` and :attr:`nz` cells of thickness almost
        equals above :code:`hc` and wider under, the strecht parameter being defined by
        :code:`theta`.

        Parameters
        ----------
        nz : int
            Number of cells.
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.
        hc : float, positive
            Reference depth :math:`[\text m]`.
        theta : float, default=6.5
            Stretching parameter toward the surface :math:`[\text{dimensionless}]`.

        Returns
        -------
        grid : Grid
            The analytic grid.
        """
        sc_w = jnp.linspace(-1, 0, nz+1)
        sc_r = (sc_w[:-1] + sc_w[1:])/2
        cs_r = (1-jnp.cosh(theta*sc_r))/(jnp.cosh(theta)-1)
        cs_w = (1-jnp.cosh(theta*sc_w))/(jnp.cosh(theta)-1)
        zw = (hc*sc_w + hbot*cs_w) * hbot/(hbot+hc)
        zr = (hc*sc_r + hbot*cs_r) * hbot/(hbot+hc)
        return cls(zr, zw)

    @classmethod
    def orca75(cls, hbot: float) -> Grid:
        r"""
        Creates the ORCA75 grid from NEMO.

        The whole grid is created then levels between the depth :attr:`hbot` and :math:`0` are
        extracted.

        Parameters
        ----------
        hbot : float, positive
            Depth of the water column :math:`[\text m]`.

        Returns
        -------
        grid : Grid
            The ORCA75 grid.
        """
        nz_orca = 75
        zsur = -3958.95137127683
        za2 = 100.7609285
        za0 = 103.9530096
        za1 = 2.415951269
        zkth = 15.3510137
        zkth2 = 48.02989372
        zacr = 7.
        zacr2 = 13.
        sc_r = jnp.arange(nz_orca-0.5, 0.5, -1)
        sc_w = jnp.arange(nz_orca, 0, -1)
        zw_orca = -(zsur + za0*sc_w + za1*zacr*jnp.log(jnp.cosh((sc_w-zkth)/zacr)) +\
                    za2*zacr2*jnp.log(jnp.cosh((sc_w-zkth2)/zacr2)))
        zr_orca = -(zsur + za0*sc_r + za1*zacr*jnp.log(jnp.cosh((sc_r-zkth)/zacr)) +\
                    za2*zacr2*jnp.log(jnp.cosh((sc_r-zkth2)/zacr2)))
        ibot = jnp.argmin(zw_orca <= -hbot)
        if ibot == 0:
            ibot = 1
        zw_orca = zw_orca.at[-1].set(0.)
        zw = zw_orca[ibot-1:]
        zr = zr_orca[ibot-1:]
        return cls(zr, zw)

    @classmethod
    def load(cls, ds: xr.Dataset) -> Grid:
        """
        Creates the grid defined by a dataset :code:`ds` of an observation.

        The dataset must be formated to have the variables corresponding to :attr:`zr` and
        :attr:`zw`.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset from which to extract the grid.

        Returns
        -------
        grid : Grid
            The loaded grid.
        """
        zw = jnp.array(ds['zw'], dtype=jnp.float32)
        zr = jnp.array(ds['zr'], dtype=jnp.float32)
        return cls(zr, zw)


class State(eqx.Module):
    r"""
    Water column state at one time-step.

    This state is defined on a :attr:`grid` describing the geometry, and is composed of the
    variables of the water column : the values of the momentum variables :attr:`u` and :attr:`v`
    (which are mandatory) and the tracers variables :attr:`t`, :attr:`s`, :attr:`b` and :attr:`pt`
    (which are optionals). The constructor takes all the attributes as parameters.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column.
    u : float :class:`~jax.Array` of shape (nz)
        Zonal velocity on the center of the cells :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v : float :class:`~jax.Array` of shape (nz)
        Meridional velocity on the center of the cells :math:`\left[\text m \cdot
        \text s^{-1}\right]`.
    t : float :class:`~jax.Array` of shape (nz), optionnal, default=None
        Temperature on the center of the cells :math:`[° \text C]`.
    s : float :class:`~jax.Array` of shape (nz), optionnal, default=None
        Salinity on the center of the cells :math:`[\text{psu}]`.
    b : float :class:`~jax.Array` of shape (nz), optionnal, default=None
        Buoyancy on the center of the cells :math:`[\text{dimensionless}]`.
    pt : float :class:`~jax.Array` of shape (nz), optionnal, default=None
        A passive tracer on the center of the cells :math:`[\text{dimensionless}]`.

    """

    grid: Grid
    u: ArrNz
    v: ArrNz
    t: Optional[ArrNz] = None
    s: Optional[ArrNz] = None
    b: Optional[ArrNz] = None
    pt: Optional[ArrNz] = None

    @classmethod
    def zeros(cls, grid: Grid, tracers: List[str]) -> State:
        """
        Initialize an instance with all variables equals to zero from a grid.

        Parameters
        ----------
        grid : Grid
            Geometry of the water column.

        Returns
        -------
        state : State
            An instance defined on the grid with all variables set to 0.        
        """
        zero_array = jnp.zeros(grid.nz)
        tracers_dict = {}
        for tracer_name in tracers:
            tracers_dict[tracer_name] = zero_array
        return State(grid, u=zero_array, v=zero_array, **tracers_dict)

    def init_u(self, hmxl: float=20., u_sfc: float=0.) -> State:
        r"""
        Initialize zonal velocity with a classical wind stratification.

        Return a State object where :attr:`u` is continuous and linear by part
        :math:`u(z) = \begin{cases}
        0 & \text{if } z < h_{\text{mxl}}\\
        u_{\text{sfc}} \left( 1 - \dfrac z {h_{\text{mxl}}}\right) &
        \text{else} \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        u_sfc : float, default=0.
            Surface zonal velocity :math:`\left[\text m \cdot \text s^{-1}\right]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of zonal velocity.
        """
        maped_fun = vmap(_piecewise_linear_ramp, in_axes=(0, None, None))
        maped_fun = cast(Callable[[ArrNz, float, float], ArrNz], maped_fun)
        u_new = maped_fun(self.grid.zr, -hmxl, u_sfc)
        return eqx.tree_at(lambda t: t.u, self, u_new)

    def init_v(self, hmxl: float=20., v_sfc: float=0.) -> State:
        r"""
        Initialize meridional velocity with a classical wind stratification.

        Return a State object where :attr:`v` is continuous and linear by part
        :math:`v(z) = \begin{cases}
        0 & \text{if } z < h_{\text{mxl}}\\
        v_{\text{sfc}} \left( 1 - \dfrac z {h_{\text{mxl}}}\right) &
        \text{else} \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        u_sfc : float, default=0.
            Surface meridional velocity :math:`\left[\text m \cdot \text s^{-1}\right]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of meridional velocity.
        """
        maped_fun = vmap(_piecewise_linear_ramp, in_axes=(0, None, None))
        maped_fun = cast(Callable[[ArrNz, float, float], ArrNz], maped_fun)
        v_new = maped_fun(self.grid.zr, -hmxl, v_sfc)
        return eqx.tree_at(lambda t: t.v, self, v_new)

    def init_t(self, hmxl: float=20., t_sfc: float=21., strat_t: float=5.1e-2) -> State:
        r"""
        Initialize temperature with a classical tracer stratification.

        Return a State object where :attr:`t` is linear by part and continous
        :math:`T(z) = \begin{cases}
        t_{\text{sfc}} + S_T(z-h_{\text{mxl}}) & \text{if } z <
        h_{\text{mxl}}\\
        t_{\text{sfc}} & \text{else}
        \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        t_sfc : float, default=21.
            Surface temperature :math:`[° \text C]`.
        strat_t : float, default=5.1e-2
            Thermal stratification above the mixed layer noted by :math:`S_T`
            :math:`[\text K \cdot \text m ^{-1}]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of temperature.
        """
        maped_fun = vmap(_piecewise_linear_flat, in_axes=(0, None, None, None))
        maped_fun = cast(Callable[[ArrNz, float, float, float], ArrNz], maped_fun)
        t_new = maped_fun(self.grid.zr, -hmxl, t_sfc, strat_t)
        return eqx.tree_at(lambda tree: tree.t, self, t_new)

    def init_s(self, hmxl: float=20., s_sfc: float=35., strat_s: float=1.3e-2) -> State:
        r"""
        Initialize salinity with a classical tracer stratification.

        Return a State object where :attr:`s` is linear by part and continous
        :math:`S(z) = \begin{cases}
        s_{\text{sfc}} + S_S(z-h_{\text{mxl}}) & \text{if } z <
        h_{\text{mxl}}\\
        s_{\text{sfc}} & \text{else}
        \end{cases}`

        Parameters
        ----------
        hmxl : float, default=20.
            Mixed layer depth :math:`[\text m]`.
        s_sfc : float, default=21.
            Surface salinity :math:`[\text{psu}]`.
        strat_s : float, default=5.1e-2
            Salinity stratification above the mixed layer noted by :math:`S_T`
            :math:`[\text{psu} \cdot \text m ^{-1}]`.

        Returns
        -------
        state : State
            The :code:`self` object with the the new value of temperature.
        """
        maped_fun = vmap(_piecewise_linear_flat, in_axes=(0, None, None, None))
        maped_fun = cast(Callable[[ArrNz, float, float, float], ArrNz], maped_fun)
        s_new = maped_fun(self.grid.zr, -hmxl, s_sfc, strat_s)
        return eqx.tree_at(lambda t: t.s, self, s_new)


class Trajectory(eqx.Module):
    r"""
    Define the history of a simulation or an observation.

    Contains the timeseries of the momentum (mandatory) variables and the tracers variables
    (optionals) throught the space of the :attr:`grid` and the :attr:`time`. The constructor takes
    all the attributes as parameters.

    Attributes
    ----------
    grid : Grid
        Geometry of the water column.
    time : float :class:`~jax.Array` of shape (nt)
        Time at each steps of observation from the begining of the simulation :math:`[\text s]`.
    u : float :class:`~jax.Array` of shape (nz, nt)
        Time-serie of zonal velocity :math:`\left[\text m \cdot \text s^{-1}\right]`.
    v : float :class:`~jax.Array` of shape (nz, nt)
        Time-serie of meridional velocity :math:`\left[\text m \cdot \text s^{-1}\right]`.
    t : float :class:`~jax.Array` of shape (nz, nt), optionnal, default=None
        Time-serie of temperature :math:`[\text C°]`.
    s : float :class:`~jax.Array` of shape (nz, nt), optionnal, default=None
        Time-serie of salinity :math:`[\text{psu}]`.
    b : float :class:`~jax.Array` of shape (nz, nt), optionnal, default=None
        Time-serie of buoyancy :math:`[\text{dimensionless}]`.
    pt : float :class:`~jax.Array` of shape (nz, nt), optionnal, default=None
        Time-serie a passive tracer :math:`[\text{dimensionless}]`.
        
    """

    grid: Grid
    time: Float[Array, 'nt']
    u: ArrNzNt
    v: ArrNzNt
    t: Optional[ArrNzNt] = None
    s: Optional[ArrNzNt] = None
    b: Optional[ArrNzNt] = None
    pt: Optional[ArrNzNt] = None

    def to_ds(self) -> xr.Dataset:
        """
        Exports the trajectory in an xarray.Dataset.

        The dimensions of the dataset are :attr:`time`, :code:`grid.zr` and :code:`grid.zw`, the
        variables are :attr:`u`, :attr:`v` and the tracers that are not set to :code:`None`, all
        defined on the dimensions (:attr:`time`, :code:`zr`) or :code:`zw` depending on
        :data:`VARIABLE_NAMES`.

        Returns
        -------
        ds : xarray.Dataset
            Dataset of the trajectory.
        """
        variables = {}
        for var_name in VARIABLE_NAMES:
            var = getattr(self, var_name)
            if var is not None:
                variables[var_name] = (('time', VARIABLE_SHAPES[var_name]), var)
        coords = {
            'time': self.time,
            'zr': self.grid.zr,
            'zw': self.grid.zw
        }
        return xr.Dataset(variables, coords)

    def to_nc(self, nc_path: str) -> None:
        r"""
        Write on a NetCDF file.

        The dimensions are :attr:`time`, :code:`grid.zr` and :code:`grid.zw`, the variables are
        :attr:`u`, :attr:`v` and the tracers that are not set to :code:`None`, all defined on the
        dimensions (:attr:`time`, :code:`zr`) or :code:`zw` depending on :data:`VARIABLE_NAMES`.
        
        Parameters
        ----------
        nc_path : str
            Path of the file on which write the trajectory.
        """
        ds = self.to_ds()
        ds.to_netcdf(nc_path)

    def extract_state(self, i_time: int) -> State:
        """
        Extracts the water column state at one time index.

        Parameters
        ----------
        i_time : int
            The time index of the moment to extract the state.

        Returns
        -------
        state : State
            The state of the trajectory at the time of index :code:`i_time`.
        """
        variables = {}
        for var_name in VARIABLE_NAMES:
            var = getattr(self, var_name)
            if var is not None:
                variables[var_name] = var[i_time, :]
        return State(self.grid, **variables)

    def cut(self, out_nt_cut: int) -> List[Trajectory]:
        """
        Cut the trajectory in sub-trajectories of :code:`out_nt_cut` output steps.

        The first and last state of two consecutive trajectories are the same. :code:`out_nt_cut`
        is the number of output steps, it means that the time dimension of the sub-trajectories
        have :code:`out_nt_cut+1` elements.

        Parameters
        ----------
        out_nt_cut : int
            Number of output steps of the sub-trajectories.
        
        Returns
        -------
        traj_list : List[Trajectory]
            List of the sub-trajectories in the chronological order.

        Warns
        -----
        Lost last trajectory
            If :code:`out_nt_cut` does not divide the number of output step of the initial
            trajectory. In this case the last part of the trajectory (which is too short) is
            abandonned.
        """
        out_nt = self.time.shape[0] - 1
        if out_nt%out_nt_cut != 0:
            warnings.warn(_format_to_single_line("""
                If out_nt_cut does not divide the number of output step of the initial trajectory.
                In this case the last part of the trajectory (which is too short) is abandonned.
            """))
        traj_list = []
        n_cuts = out_nt//out_nt_cut
        for i_cut in range(n_cuts):
            i_start, i_end = out_nt_cut*i_cut, out_nt_cut*(i_cut+1) + 1
            cut_time = self.time[i_start:i_end]
            var_dict_cut = {}
            for var in VARIABLE_NAMES:
                if getattr(self, var) is not None:
                    var_dict_cut[var] = getattr(self, var)[i_start:i_end, :]
            cut_traj = Trajectory(self.grid, cut_time, **var_dict_cut)
            traj_list.append(cut_traj)
        return traj_list



"""
Abstraction usefull for the calibration of the closures.

This module contains the classes :class:`FittableParameter` and :class:`FittableParametersSet` which
are used to make a link between the optimization part and the closures structures. These class can
be obtained by the prefix :code:`tunax.fitter.` or directly by :code:`tunax.`.

"""

from __future__ import annotations
from typing import Dict

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Float, Array

from tunax.closure import ClosureParametersAbstract, Closure
from tunax.closures_registry import CLOSURES_REGISTRY


class FittableParameter(eqx.Module):
    """
    Calibration configuration for one parameter.

    An instance of this class must be created for every parameter of the
    closure that will be calibrated and for every not default values during
    calibration. The constructor takes all the attributes as parameters.

    Attributes
    ----------
    do_fit : bool
        The parameter will be calibrated.
    val : float, default=0.
        If :attr:`do_fit` : the initial value for calibration (at the first
        step of the calibration) ; if :attr:`do_fit` is false : the constant
        value to take for this parameter if it's not the default one in the
        closure.

    """

    do_fit: bool
    val: float = 0.


class FittableParametersSet(eqx.Module):
    """
    Complete closure calibration parameters.

    This class is the set of all the configurations on the closure parameters
    for the calibration. It makes the link between the array on which the
    optimizer works and the closure parameters class.

    Parameters
    ----------
    coef_fit_dict : Dict[str, FittableParameter]
        cf. :attr:`coef_fit_dict`.
    closure_name : str
        Name of the chosen closure, must be a key of
        :attr:`~closures_registry.CLOSURES_REGISTRY`, see its documentation
        for the available closures.

    Attributes
    ----------
    coef_fit_dico : Dict[str, FittableParameter]
        The set of all the configurations of all the parameters that will be
        calibrated and the one constants but not with the default value of the
        closure.
    closure : Closure
        The abstraction that represent the used closure.
    
    """

    coef_fit_dict: Dict[str, FittableParameter]
    closure: Closure

    def __init__(
            self,
            coef_fit_dict: Dict[str, FittableParameter],
            closure_name: str
        ) -> None:
        self.coef_fit_dict = coef_fit_dict
        self.closure = CLOSURES_REGISTRY[closure_name]

    @property
    def n_calib(self) -> int:
        """
        Number of variables that are calibrated.

        Returns
        -------
        nc : int
            Number of variables that are calibrated.
        """
        nc = 0
        for coef_fit in self.coef_fit_dict.values():
            if coef_fit.do_fit:
                nc += 1
        return nc

    def fit_to_closure(
            self,
            x: Float[Array, 'nc']
        ) -> ClosureParametersAbstract:
        """
        Transforms an fitted array in a set of closure parameters.

        This method copy the fixed non-default values of :attr:`coef_fit_dict`
        and copy the calibrated values from :code:`x`. Which is simply the
        parameters values in the order that is indicated by
        :attr:`coef_fit_dict`.

        Parameters
        ----------
        x : float :class:`~jax.Array` of shape (nc)
            The array on which the optimize works to find the best values. It
            is the array of the parameters that are calibrated.

        Returns
        -------
        clo_params : ClosureParametersAbstract
            The instance of the closure parameters class (child class of
            :class:`~closure.ClosureParametersAbstract`) with the modifications
            of the calibration step.
        """
        clo_coef_dico = {}
        i_x = 0
        for coef_name, coef_fit in self.coef_fit_dict.items():
            if coef_fit.do_fit:
                clo_coef_dico[coef_name] = x[i_x]
                i_x += 1
            else:
                clo_coef_dico[coef_name] = coef_fit.val
        return self.closure.parameters_class(**clo_coef_dico)

    def gen_init_val(self) -> Float[Array, 'nc']:
        """
        Produce the fitted array for the first calibration step.

        This method simply copy the initial values of the calibrated
        coefficients in an array :code:`x` which will be used as the first
        calibration step for the optimizer.

        Returns
        -------
        x : float :class:`~jax.Array` of shape (nc)
            The initial vector for the optimizer at the first step of
            calibration.
        """
        x = []
        for coef_fit in self.coef_fit_dict.values():
            if coef_fit.do_fit:
                x.append(coef_fit.val)
        return jnp.array(x)



"""
Abstraction for calibration databases.


This module the objects that are used in Tunax to describe a :class:`Database` of observations used
for a calibration. By *datas* (:class:`Data`), we refer ton the union of a trajectory and a physical
case which represent a measurment or a reference as a Large Eddy Simulation (LES) for example. And
by *observations* (:class:`Obs`), we refer to the union of a trajectory and the Tunax model which
corresponds to it. These classes can be obtained by the prefix :code:`tunax.database.` or directly
by :code:`tunax.`.

"""

from __future__ import annotations
import warnings
from typing import Union, Optional, Tuple, List, Dict, TypeAlias

import yaml
import xarray as xr
import equinox as eqx
import numpy as np
import jax.numpy as jnp
from h5py import File as H5pyFile
from h5py import Dataset as H5pyDataset
from h5py import Group as H5pyGroup
from jaxtyping import Array, Float

from tunax.space import Grid, Trajectory, TRACERS_NAMES, VARIABLE_NAMES
from tunax.case import Case
from tunax.functions import _format_to_single_line
from tunax.model import SingleColumnModel

DimsType: TypeAlias = Tuple[Optional[int]]
"""Type that represent the dimensions on which load the datas from a file."""


def _get_var_jl(
        jl_file: H5pyFile,
        var_names: Dict[str, str],
        var: str,
        n: int,
        dims: Union[DimsType, Dict[str, DimsType]] = (None,),
        suffix: str = ''
    ) -> Float[Array, 'n']:
    """
    This function retrieves the right value of a variable in a .jld2 file.

    This function retrives first the raw data that corresponding to the variable var (name from
    Tunax) in accord to the names registry var_names, with an eventual suffix at the end. In a
    second time it will project in the good dimensions the raw data with the indications in dims to
    get a one-dimension array. And finally it will remove the eventual borders by taking the middle
    part of the array of lenght n.
    
    Parameters
    ----------
    jl_file : H5pyFile
        A .jld2 file loaded with the package h5py.
    var_names : Dict[str, str]
        The reference on which search the variable on the file, the keys are the Tunax names of
        variables and the values are the names in the file (which are actually paths).
    var : str
        The name of the variable to search in terms of Tunax.
    n : int
        The excpected lenght of the variable. The function will keep the middle part of lenght n
        on the array that it will directly extract from the file.
    dims : DimsType or Dict[str, DimsType], default=(None,)
        It contains the dimensions on which search the right array. If it's a dictionnary, it's like
        var_names, the keys are the names of the variables in terms of Tunax and the values are the
        dimensions for each variable. Then we have a Tuple of int or Nones which corresponds at
        every axis of the raw data from the file. If an axis is indexed with None, it means that we
        keep this dimension, if an axis is indexed with an integer, it means that we reduce this
        axis to the value of the raw data on this index.
    suffix : str, default=''
        A string suffix to add at the end of the path that is in var_names.

    Returns
    -------
    arr : float :class:`~jax.Array` of shape (n)
        Value of the right array in the :code:`.jld2` file.
    
    Raises
    ------
    ValueError
        If the lenght of the tuple dims doesn't have the same number of elements than the number of
        axis in the raw data from the file.

    Warns
    -----
    Odd shift
        If the number n hasn't the same parity as the lenght of the raw data array. Then the borders
        are not symetric.
    """
    jl_var = H5pyDataset(jl_file[f'{var_names[var]}{suffix}'])
    if isinstance(dims, dict):
        dims = dims[var]
    if len(jl_var.shape) != len(dims):
        raise ValueError(_format_to_single_line(f"""
            The tuple parameter `dims` of value {dims} must have the length of the number of
            dimension of the variable {var} array in the `jl_file`.
        """))
    dims_slice = tuple(slice(None) if x is None else x for x in dims)
    jl_var_1d = jl_var[dims_slice]
    double_shift = jl_var_1d.shape[0] - n
    shift = double_shift//2
    if double_shift%2 == 1:
        warnings.warn(_format_to_single_line(f"""
            The length array from the `jl_file` of the variable {var} minus `n` is an odd number :
            the removed boundaries are taken 1 point thinner on the bottom side than on the surface
            side.
        """))
        return jl_var_1d[shift:-shift-1]
    return jl_var_1d[shift:-shift]


class Data(eqx.Module):
    """
    Abstraction to represent an element of the database from the point of view of Tunax.

    This abstraction is the link between the time-series of :class:`Trajectory` and a physical
    situation described by :class:`Case`. It can eventually contains metadatas. Typically this class
    appears when one want to import the different element of a database of observations or
    simulations. The constructor takes all the attributes as parameters.
    
    Attributes
    ----------
    trajectory : Trajectory
        The time-series of the variables that represent this data.
    case : Case
        The physical case that represent this data.
    metadatas : Dict[str, float], default={}
        Some metadatas that we want to use later. It can be some values of the :attr:`case` that
        we want to set by hand later.

    Raises
    ------
    ValueError
        If the :attr:`~space.Trajectory.time` of :attr:`trajectory` is not build with constant
        time-steps.

    """

    trajectory: Trajectory
    case: Case
    metadatas: Dict[str, float] = eqx.field(static=True)

    def __init__(
            self,
            trajectory: Trajectory,
            case: Case,
            metadatas: Optional[Dict[str, float]]=None
        ) -> None:
        time = trajectory.time
        steps = time[1:] - time[:-1]
        if not jnp.all(steps == steps[0]):
            raise ValueError('Tunax only handle constant output time-steps')
        self.trajectory = trajectory
        self.case = case
        if metadatas is None:
            metadatas = {}
        self.metadatas = metadatas

    @classmethod
    def from_nc_yaml(
            cls,
            nc_path: str,
            yaml_path: str,
            var_names: Dict[str, str],
            eos_tracers: str = 't',
            do_pt: bool = False
        ) -> Data:
        """
        Create a :class:`Data` instance from a *netcdf* and a *yaml* files.

        This class method build a trajectory from the :code:`.nc` file :code:`nc_path`, it build the
        physical parameters from the configuration file :code:`yaml_path`. :code:`var_names` is used
        to do the link between Tunax name convention and the one from the used database.

        Parameters
        ----------
        nc_path : str
            Path of the *netcdf* file that contains the time-series of the observation trajectory.
            The file should contains at least the three dimensions :attr:`~space.Grid.zr`,
            :attr:`~space.Grid.zw` and :attr:`~space.Trajectory.time`. The time-series can be
            created with default values if they are not present in the file (only for
            :attr:`space.Trajectory.u` and :attr:`space.Trajectory.v`). Otherwise, they must have
            the good dimensions described in :class:`~space.Trajectory`.
        yaml_path : str
            Path of the *yaml* file that contains the parameters and forcing that describe the
            observation. The parameters should be float numbers and directly accessible from the
            root of the file with a key. Only the parameters that are described in
            :class:`~case.Case` will be takend in account.
        var_names : Dict[str, str]
            Link between the convention names in Tunax and the ones in the database. The keys are
            the Tunax names and the values are the names in the database. It works for variables of
            the :class:`~space.Trajectory` and fornthe parameters of :class:`~case.Case`. It must at
            least contains entries for :attr:`~space.Grid.zr` :attr:`~space.Grid.zw` and
            :attr:`~space.Trajectory.time`
        eos_tracers : str, default='t'
            Tracers used for the equation of state, cf. :attr:`~case.Case.eos_tracers`.
        do_pt : bool, default=False
            Compute or not a passive tracer, cf. :attr:`~case.Case.do_pt`.
        
        Returns
        -------
        data : Data
            An object that represent these files.
        """
        ds = xr.load_dataset(nc_path)
        # dimensions
        zr = jnp.array(ds[var_names['zr']].values)
        zw = jnp.array(ds[var_names['zw']].values)
        grid = Grid(zr, zw)
        time = jnp.array(ds[var_names['time']].values)
        nt, = time.shape
        nz = grid.nz
        # variables
        u_name = var_names['u']
        if u_name == '':
            u = jnp.full((nt, nz), 0.)
        else:
            u = jnp.array(ds[u_name].values)
        v_name = var_names['v']
        if v_name == '':
            v = jnp.full((nt, nz), 0.)
        else:
            v = jnp.array(ds[v_name].values)
        tracers_dict = {}
        for tracer in TRACERS_NAMES:
            if tracer in var_names.keys():
                tracers_dict[tracer] = jnp.array(ds[var_names[tracer]].values)
        # writing trajectory
        trajectory = Trajectory(grid, time, u, v, **tracers_dict)
        # writing the case
        with open(yaml_path, 'r', encoding='utf-8') as f:
            metadatas = yaml.safe_load(f)
        case = Case()
        case_attributes = list(vars(case).keys())
        def get_pytree_fun(att: str):
            return lambda t: getattr(t, att)
        for att in case_attributes:
            if att in var_names.keys():
                case = eqx.tree_at(get_pytree_fun(att), case, metadatas[var_names[att]])
        case = eqx.tree_at(lambda t: getattr(t, 'eos_tracers'), case, eos_tracers)
        case = eqx.tree_at(lambda t: getattr(t, 'do_pt'), case, do_pt)

        return cls(trajectory, case, metadatas)

    @classmethod
    def from_jld2(
            cls,
            jld2_path: str,
            names_mapping: Dict[str, Dict[str, str]],
            nz: Optional[int] = None,
            dims: Union[DimsType, Dict[str, DimsType]] = (None,),
            eos_tracers: str = 't',
            do_pt: bool = False
        ) -> Data:
        """
        Creates a :class:`Data` instance from a :code:`.jld2` file.

        For the scalar parameters, the values must be registered in the file simply with their path
        in the file, separated with :code:`/` in the same string. For the timeseries and the time,
        the arrays of each time step must be register with a path that ends with the reference of
        the time. For the other variables, just register the normal path.

        Parameters
        ----------
        jld2_path : str
            Path of the *netcdf* file that contains the time-series of the observation trajectory
            and the physical parameters and forcings.
        names_mapping: Dict[str, Dict[str, str]]
            Contains the link between the Tunax names of variables and the path of the variables in
            the file. There are 3 first entries :

            - :code:`variables` :for all the variables corresponding to the :class:`space.Grid` and
              the :class:`space.Trajectory`. For the grid attributes (:attr:`space.Grid.zr` and
              :attr:`space.Grid.zw`) the path should correspond directly to the array in the file.
              For the time and the time-series, the given path corresponds to a path with all the
              reference (with a number in string) of the time, and then in these path with the time
              we have the array of the variable (or the float corresponding to the value of the
              time) at the time with this reference. Then the 2D arrays are rebuild by
              concatenation. The references of the time are get with the path of the time data.

            - :code:`parameters` : for all the scalar entries corresponding directly to the
              parameters of :class:`case.Case`

            - :code:`metadatas` : for the scalar entries that we want to keep in the
              :attr:`metadatas` for later.
              
        nz : int, optionnal, default=None
            Expected number of steps of the grid of the water column. The method will remove the
            borders of the raw data from the file to keep only the middle part of this lenght. If
            nothing is entered for this parameter, all the raw data are kept.
        dims : DimsType or Dict[str, DimsType], default=(None,)
            It contains the dimensions on which search the right arrays. If it's a dictionnary,
            it's like :code:`var_names`, the keys are the names of the variables in terms of Tunax
            and the values are the dimensions for each variable. Then we have a Tuple of int or
            Nones which corresponds at every axis of the raw data from the file. If an axis is
            indexed with None, it means that we keep this dimension, if an axis is indexed with an
            integer, it means that we reduce this axis to the value of the raw data on this index.
        eos_tracers : str, default='t'
            Tracers used for the equation of state, cf. :attr:`~case.Case.eos_tracers`.
        do_pt : bool, default=False
            Compute or not a passive tracer, cf. :attr:`~case.Case.do_pt`.
        
        Returns
        -------
        data : Data
            An object that represent these file.
        """
        var_map = names_mapping['variables']
        par_map = names_mapping['parameters']
        metadata_map = names_mapping['metadatas']
        jl = H5pyFile(jld2_path, 'r')
        # récupération de la bonne valeur de nz
        if nz is None:
            ds = H5pyDataset(jl[par_map['nz']])
            nz = int(ds[()])
        # variables grid et time
        zr = jnp.array(_get_var_jl(jl, var_map, 'zr', nz, dims))
        zw = jnp.array(_get_var_jl(jl, var_map, 'zw', nz+1, dims))
        time_group = var_map['time']
        gr = H5pyGroup(jl[time_group])
        time_str_list = list(gr.keys())
        time_str_list = [int(i) for i in time_str_list]
        time_str_list.sort()
        time_str_list = [str(i) for i in time_str_list]
        time_float_list = []
        for time_str in time_str_list:
            ds = H5pyDataset(jl[f'{time_group}/{time_str}'])
            time_val = ds[()]
            time_float_list.append(float(time_val))
        time = jnp.array(time_float_list)
        # variables
        variables_dict = {}
        for var_name in VARIABLE_NAMES:
            if var_name not in var_map:
                continue
            var_list = []
            for time_str in time_str_list:
                var_time = _get_var_jl(jl, var_map, var_name, nz, dims, f'/{time_str}')
                var_list.append(var_time)
            variables_dict[var_name] = jnp.vstack(var_list)
        # trajectory
        trajectory = Trajectory(Grid(zr, zw), time, **variables_dict)

        # parameters
        params = {}
        case_params_list = [nom for nom in vars(Case).keys()]
        for par_name, jl_name in par_map.items():
            if par_name in case_params_list:
                ds = H5pyDataset(jl[jl_name])
                params[par_name] = float(ds[()])
        case = Case(eos_tracers=eos_tracers, do_pt=do_pt, **params)

        # metadatas
        metadatas = {}
        for metadata_name, jl_name in metadata_map.items():
            ds = H5pyDataset(jl[jl_name])
            jl_val = ds[()]
            if isinstance(jl_val, np.floating) or isinstance(jl_val, np.integer):
                metadatas[metadata_name] = float(jl_val)

        return cls(trajectory, case, metadatas)

    def cut(self, out_nt_cut: int) -> List[Data]:
        """
        Cuts the :attr:`Trajectory` in sub-trajectories, cf. :meth:`space.Trajectory.cut`.

        Parameters
        ----------
        out_nt_cut : int
            Number of output steps of the sub-trajectories.
        
        Returns
        -------
        traj_list : List[Data]
            List of :class:`Data` instances with the sub-trajectories in the chronological order.
        """
        traj_list = self.trajectory.cut(out_nt_cut)
        return [Data(traj, self.case, self.metadatas) for traj in traj_list]


class Weights(eqx.Module):
    """
    Representation of the weights to put on every variable for the computing of the lost function.
    The constructor takes all the attributes as parameters.

    Attributes
    ----------
    weight_u : float, default=0.
        Weight on zonal velocity.
    weight_v : float, default=0.
        Weight on meridionnal velocity.
    weight_t : float, default=0.
        Weight on temperature.
    weight_s : float, default=0.
        Weight on salinity.
    weight_b : float, default=0.
        Weight on buoyancy.
    weight_pt : float, default=0.
        Weight on passive tracer.
    """
    weight_u: float = 0.
    weight_v: float = 0.
    weight_t: float = 0.
    weight_s: float = 0.
    weight_b: float = 0.
    weight_pt: float = 0.


class Obs(eqx.Module):
    """
    This class represents and element of the database from the point of view of the loss function.
     
    This class prepares everything to make the loss function able to compute the loss for this
    element of the database. Indeed this class makes the link between the :class:`Trajectory`
    corresponding to this element, a model (with a grid a time parameters) corresponding to this
    trajectory, and the weights that we want to put on each variable. The constructor takes all the
    attributes as parameters.

    Attributes
    ----------
    trajectory: Trajectory
        The time-series of the variables that represent this observation.
    model: SingleColumnModel
        A model built on this trajectory and on a physical case with the time and geometrical
        parameters.
    weights: Weights
        The weights to give to the loss function.
    """
    trajectory: Trajectory
    model: SingleColumnModel
    weights: Weights

    @classmethod
    def from_data(cls, data: Data, dt: float, weights: Weights):
        """
        Create a Obs instance from a Data one adding :class:`Weights` and a :code:`dt`.

        This function builds the other time parameters of the model from the trajectory.

        Parameters
        ----------
        data : Data
            A data containing the trajectory and the physical case that we want to apply on our
            model.
        dt : float
            The integration time-step that we want for our model.
        weights : Weights
            The weights to give to the loss function.
        """
        time = data.trajectory.time
        nt = int((time[-1]-time[0])/dt)
        out_dt = float(time[1] - time[0])
        p_out = int(out_dt/dt)
        init_state = data.trajectory.extract_state(0)
        start_time = float(time[0])
        model = SingleColumnModel(nt, dt, p_out, init_state, data.case, 'k-epsilon', start_time)
        return Obs(data.trajectory, model, weights)


class Database(eqx.Module):
    """
    Represent a set of several observations that form a database. The constructor takes all the
    attributes as parameters.

    Attributes
    ----------
    observations : List[Obs]
        A list of several observations with potentially various forcings, geometry and time
        configuration.

    """

    observations: List[Obs]
